#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`589`-compliant **typed dictionary** (i.e.,
:class:`typing.TypedDict` subclass) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._util.cls.utilclstest import is_type_subclass
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14

# ....................{ TESTERS                            }....................
def is_hint_pep589(hint: object) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`589`-compliant **typed
    dictionary** (i.e., :class:`typing.TypedDict` subclass).

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to a
    one-liner. Moreover, callers are *not* expected to call this tester
    directly. Instead, callers are expected to (in order):

    #. Call the memoized
       :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign_or_none`
       getter, which internally calls this unmemoized tester.
    #. Compare the object returned by that getter against the
       :attr:`beartype._util.data.hint.pep.sign.datahintsigns.HintSignTypedDict`
       sign.

    Motivation
    ----------
    The implementation of the :obj:`typing.TypedDict` attribute substantially
    varies across Python interpreter *and* :mod:`typing` implementation.
    Notably, the :obj:`typing.TypedDict` attribute under Python >= 3.9 is *not*
    actually a superclass but instead a factory function masquerading as a
    superclass by setting the subversive ``__mro_entries__`` dunder attribute to
    a tuple containing a private :obj:`typing._TypedDict` superclass. That
    superclass then defines various requisite dunder attributes. Passing the
    passed hint with that superclass to the :func:`issubclass` builtin fails, as
    the metaclass of that superclass prohibits :func:`issubclass` checks. I am
    throwing up in my mouth as I write this.

    Unfortunately, all of the above complications are further complicated by the
    :class:`dict` type under Python >= 3.10. For unknown reasons, Python >= 3.10
    adds spurious ``__annotations__`` dunder attributes to *all* :class:`dict`
    subclasses -- even if those subclasses annotate *no* class or instance
    variables. While a likely bug, we have little choice but to at least
    temporarily support this insanity.

    Parameters
    ----------
    hint : object
        Object to be tested.

    Returns
    -------
    bool
        :data:`True` only if this object is a typed dictionary.
    '''

    # Return true only...
    #
    # Note that:
    # * This detection scheme is technically susceptible to false positives in
    #   unlikely edge cases. Specifically, this test queries for dunder
    #   attributes and thus erroneously returns true for user-defined "dict"
    #   subclasses *NOT* subclassing the "typing.TypedDict" superclass but
    #   nonetheless declaring the same dunder attributes declared by that
    #   superclass. Since the likelihood of any user-defined "dict" subclass
    #   accidentally defining these attributes is vanishingly small *AND* since
    #   "typing.TypedDict" usage is sorta discouraged in the typing community,
    #   this error is unlikely to meaningfully arise in real-world use cases.
    #   Ergo, it is preferable to implement this test portably, safely, and
    #   efficiently rather than accommodate this error.
    # * This detection scheme would ideally trivially reduce to this one-liner:
    #       from typing import is_typeddict
    #       return is_typeddict(hint)
    #
    #   Deferring to that official tester suffices for typed dictionaries
    #   defined with the official "typing.TypedDict" but *NOT* third-party
    #   "typing_extensions.TypedDict" superclass. Even that constraint could be
    #   circumvented by noting that the is_typeddict() tester itself trivially
    #   reduces to a similar one-liner:
    #       return isinstance(tp, _TypedDictMeta)
    #
    #   An alternate working solution covering both the "typing.TypedDict" and
    #   "typing_extensions.TypedDict" superclasses would then resemble:
    #       _TYPED_DICT_BASES = (
    #            typing._TypedDictMeta, typing_extensions._TypedDictMeta,)
    #       return isinstance(tp, _TYPED_DICT_BASES)
    #
    #   Of course, that solution would *STILL* violate privacy encapsulation
    #   across not one but two distinct APIs, inviting future breakage.
    #
    # In short, the current approach is strongly preferable.
    return (
        # This hint is a "dict" subclass *AND*...
        #
        # If this hint is *NOT* a "dict" subclass, this hint *CANNOT* be a typed
        # dictionary. By definition, typed dictionaries are "dict" subclasses.
        # Note that PEP 589 actually lies about the type of typed dictionaries:
        #     Methods are not allowed, since the runtime type of a TypedDict
        #     object will always be just dict (it is never a subclass of dict).
        #
        # This is *ABSOLUTELY* untrue. PEP 589 authors plainly forgot to
        # implement this constraint. Contrary to the above:
        # * All typed dictionaries are subclasses of "dict".
        # * The type of typed dictionaries is the private
        #   "typing._TypedDictMeta" metaclass across all Python versions (as of
        #   this comment). Although this metaclass *COULD* currently be used to
        #   detect "TypedDict" subclasses, doing so would increase the fragility
        #   of this fragile codebase by unnecessarily violating privacy
        #   encapsulation.
        #
        # This is where we generously and repeatedly facepalm ourselves.
        is_type_subclass(hint, dict) and
        # The set intersection of the set of the names of all class attributes
        # of this "dict" subclass with the frozen set of the names of all class
        # attributes unique to all "TypedDict" subclasses produces a new set
        # whose length is exactly that of the latter frozen set. Although
        # cumbersome to cognitively parse, this test is *REALLY* fast.
        len(hint.__dict__.keys() & _TYPED_DICT_UNIQUE_ATTR_NAMES) == (
            _TYPED_DICT_UNIQUE_ATTR_NAMES_LEN)
    )

# ....................{ PRIVATE ~ globals                  }....................
# The is_hint_pep589() tester defined above uniquely identifies "TypedDict"
# subclasses as types declaring *ALL* of:
_TYPED_DICT_UNIQUE_ATTR_NAMES = frozenset((
    (
        # If the active Python interpreter targets Python >= 3.14, this PEP
        # 649-compliant dunder attribute. CPython now dynamically creates the
        # "__annotations__" dunder dictionary on first access. Ergo, the
        # "__annotations__" dunder dictionary is no longer guaranteed to exist
        # as an item of the "__dict__" dunder dictionary of "TypedDict" types.
        '__annotate_func__'
        if IS_PYTHON_AT_LEAST_3_14 else
        # Else, the active Python interpreter targets Python <= 3.13. In this
        # case, this PEP 484-compliant dunder attribute.
        '__annotations__'
    ),
    '__optional_keys__',
    '__required_keys__',
    '__total__',
))
'''
Frozen set of the names of all class attributes universally unique to *all*
:class:`typing.TypedDict` subclasses and thus suitable for differentiating
:class:`typing.TypedDict` subclasses from unrelated :class:`dict` subclasses.
'''


_TYPED_DICT_UNIQUE_ATTR_NAMES_LEN = len(_TYPED_DICT_UNIQUE_ATTR_NAMES)
'''
Number of class attributes unique to *all* :class:`typing.TypedDict` subclasses.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`593`-compliant **type hint test data.**
'''

# ....................{ ADDERS                             }....................
def hints_pep593_meta() -> 'List[HintPepMetadata]':
    '''
    Session-scoped fixture returning a list of :pep:`593`-compliant **type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`593`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype.typing import (
        List,
        Sequence,
        TypeVar,
        Union,
    )
    from beartype.vale import (
        Is,
        IsAttr,
        IsEqual,
        IsInstance,
        IsSubclass,
    )
    from beartype._data.hint.sign.datahintsigns import (
        HintSignAnnotated,
        HintSignList,
        HintSignUnion,
    )
    from beartype._util.api.standard.utiltyping import get_typing_attrs
    from beartype_test.a00_unit.data.data_type import (
        Class,
        Subclass,
        SubclassSubclass,
    )
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )
    from functools import partial

    # ..................{ CLASSES                            }..................
    class TruthSeeker(object):
        def __call__(self, obj: object) -> bool:
            '''
            Tester method returning :data:`True` only if the passed object
            evaluates to :data:`True` when coerced into a boolean and whose
            first parameter is ignorable.
            '''

            return bool(obj)

    # ..................{ CALLABLES                          }..................
    def is_true(ignorable_arg, obj):
        '''
        Tester function returning :data:`True` only if the passed object
        evaluates to :data:`True` when coerced into a boolean and whose first
        parameter is ignorable.
        '''

        return bool(obj)


    # Partial of the is_true() tester defined above, effectively ignoring the
    # "ignorable_arg" parameter accepted by that tester.
    is_true_partial = partial(
        is_true, 'Drank its inspiring radiance, and the wind')

    # ..................{ VALIDATORS ~ is                    }..................
    # PEP 484-compliant union of all builtin scalar types.
    Number = Union[int, float, complex]

    # Generic beartype validators defined as lambda functions.
    IsNonEmpty = Is[lambda obj: bool(obj)]

    # Numeric beartype validators defined as lambda functions.
    IsNonNegative = Is[lambda number: number >= 0]
    IsIntNonZero = Is[lambda number: isinstance(number, int) and number != 0]

    # Textual beartype validators defined as lambda functions.
    IsLengthy = Is[lambda text: len(text) > 30]
    IsSentence = Is[lambda text: text and text[-1] == '.']

    # Textual beartype validators defined as non-lambda functions.
    def _is_quoted(text):
        return '"' in text or "'" in text
    def _is_exceptional(obj):
        raise ValueError(f'Colonial. {repr(obj)}')
    IsQuoted = Is[_is_quoted]
    IsExceptional = Is[_is_exceptional]

    # Textual beartype validator synthesized from the above validators via the
    # domain-specific language (DSL) implemented by those validators.
    IsLengthyOrUnquotedSentence = IsLengthy | (IsSentence & ~IsQuoted)

    # ..................{ VALIDATORS ~ isattr                }..................
    # Arbitrary list to validate equality against below.
    AMPLY_IMPISH = ['Amply imp-ish', 'blandishments to']

    class CathecticallyEnsconceYouIn(object):
        '''
        Arbitrary class defining an arbitrary attribute whose value has *no*
        accessible attributes but satisfies a validator tested below.
        '''

        def __init__(self) -> None:
            '''
            Initialize this object by defining this attribute.
            '''

            # Initialize this attribute to a shallow copy of this list rather
            # than this list itself to properly test equality comparison.
            self.this_mobbed_triste_of = AMPLY_IMPISH[:]

    # Beartype-specific validator validating equality against that attribute.
    IsAttrThisMobbedTristeOf = IsAttr[
        'this_mobbed_triste_of', IsEqual[AMPLY_IMPISH]]

    # Instance of this class satisfying this validator.
    BOSS_EMBOSSED_ORDERING = CathecticallyEnsconceYouIn()

    # Instance of this class *NOT* satisfying this validator.
    SORDIDLY_FLABBY_WRMCASTINGS = CathecticallyEnsconceYouIn()
    SORDIDLY_FLABBY_WRMCASTINGS.this_mobbed_triste_of = [
        'An atomical caroller', 'carousing Thanatos', '(nucl‐eating',]

    # ..................{ LOCALS                             }..................
    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # ..................{ FACTORIES                          }..................
    # For each PEP-specific type hint factory importable from each currently
    # importable "typing" module...
    for Annotated in get_typing_attrs('Annotated'):
        # print(f'Exercising PEP 593 {repr(Annotated)}...')

        # ..................{ LOCALS                         }..................
        # Local variables requiring an "Annotated" type hint factory.

        # Annotated of an isinstanceable type annotated by one beartype-specific
        # validator defined as a lambda function.
        AnnotatedStrIsLength = Annotated[str, IsLengthy]

        # Validator matching any sequence of strings that is itself *NOT* a
        # string. Technically, a string is itself a sequence of characters,
        # which are themselves strings of length 1 in Python. Pragmatically, a
        # string is rarely considered to be a "sequence of strings" in the
        # common sense of that term. Validators, save us from the sins of Guido!
        SequenceNonstrOfStr = Annotated[Sequence[str], ~IsInstance[str]]

        # Type variable bounded by a beartype validator defined as a lambda
        # function.
        T_IsIntNonZero = TypeVar(
            'T_IsIntNonZero', bound=Annotated[object, IsIntNonZero])

        # Type variable bounded by two beartype validators, one defined as a
        # lambda function and one not.
        T_IsIntNonZero = TypeVar(
            'T_IsIntNonZero',
            bound=Annotated[object, IsInstance[int], IsNonEmpty],
        )

        # Type variable bounded by a PEP 604-compliant new union of beartype
        # validators defined as lambda functions.
        T_IsNumberNonNegativeOrStrNonEmpty = TypeVar(
            'T_IsNumberNonNegativeOrStrNonEmpty',
            bound=(
                Annotated[Number, IsNonNegative] |
                Annotated[str, IsNonEmpty]
            ),
        )

        # ................{ TUPLES                             }................
        # Add PEP 593-specific test type hints to this tuple global.
        hints_pep_meta.extend((
            # ..............{ ANNOTATED                          }..............
            # Annotated of an arbitrary isinstanceable type annotated by an
            # arbitrary hashable object.
            HintPepMetadata(
                hint=Annotated[str, int],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant.
                    HintPithSatisfiedMetadata(
                        'Towards a timely, wines‐enticing gate'),
                    # List of string constants.
                    HintPithUnsatisfiedMetadata([
                        'Of languished anger’s sap‐spated rushings',]),
                ),
            ),

            # Annotated of an arbitrary nested type hint.
            HintPepMetadata(
                hint=Annotated[List[str], int],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # List of string constants.
                    HintPithSatisfiedMetadata([
                        'MINERVA‐unnerving, verve‐sapping enervations',
                        'Of a magik-stoned Shinto rivery',
                    ]),
                    # String constant.
                    HintPithUnsatisfiedMetadata('Of a Spicily sated',),
                ),
            ),

            # Unhashable annotated of an isinstanceable type annotated by an
            # unhashable mutable container.
            HintPepMetadata(
                hint=Annotated[str, []],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant.
                    HintPithSatisfiedMetadata(
                        'Papally Ľust‐besmirched Merchet laws'),
                    # List of string constants.
                    HintPithUnsatisfiedMetadata([
                        "Of our ôver‐crowdedly cowed crowd's opinion‐",]),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : is          }..............
            # Annotated of the root "object" superclass annotated by a beartype
            # validator defined as a lambda function.
            HintPepMetadata(
                hint=Annotated[object, Is[
                    lambda obj: hasattr(obj, 'this_mobbed_triste_of')]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Objects defining attributes with the above name.
                    HintPithSatisfiedMetadata(BOSS_EMBOSSED_ORDERING),
                    HintPithSatisfiedMetadata(SORDIDLY_FLABBY_WRMCASTINGS),
                    # String constant.
                    HintPithUnsatisfiedMetadata(
                        'Her timid steps to gaze upon a form'),
                ),
            ),

            # Annotated of the root "object" superclass annotated by a beartype
            # validator defined as a callable object.
            HintPepMetadata(
                hint=Annotated[object, Is[TruthSeeker()]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Objects evaluating to "True" when coerced into booleans.
                    HintPithSatisfiedMetadata(True),
                    HintPithSatisfiedMetadata(
                        'Leaped in the boat, he spread his cloak aloft'),
                    # Empty string.
                    HintPithUnsatisfiedMetadata(''),
                ),
            ),

            # Annotated of the root "object" superclass annotated by a beartype
            # validator defined as a partial function.
            HintPepMetadata(
                hint=Annotated[object, Is[is_true_partial]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Objects evaluating to "True" when coerced into booleans.
                    HintPithSatisfiedMetadata(True),
                    HintPithSatisfiedMetadata(
                        'Swept strongly from the shore, blackening the waves.'),
                    # Empty string.
                    HintPithUnsatisfiedMetadata(''),
                ),
            ),

            # Annotated of an isinstanceable type annotated by a beartype
            # validator defined as a lambda function.
            HintPepMetadata(
                hint=AnnotatedStrIsLength,
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying this validator.
                    HintPithSatisfiedMetadata(
                        'To Ɯṙaith‐like‐upwreathed ligaments'),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(
                        pith=b'Down-bound',
                        # Match that the exception message raised for this
                        # object embeds the code for this validator's lambda
                        # function.
                        exception_str_match_regexes=(
                            r'Is\[.*\blen\(text\)\s*>\s*30\b.*\]',),
                    ),
                    # String constant violating this validator.
                    HintPithUnsatisfiedMetadata('To prayer'),
                ),
            ),

            # Annotated of an isinstanceable type annotated by two or more
            # beartype validators all defined as functions, specified with
            # comma-delimited list syntax.
            HintPepMetadata(
                hint=Annotated[str, IsLengthy, IsSentence, IsQuoted],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying these validators.
                    HintPithSatisfiedMetadata(
                        '"Into piezo‐electrical, dun‐dappled lights" and...'),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(
                        pith=b'Joy.',
                        # Match that the exception message raised...
                        exception_str_match_regexes=(
                            # Embeds the code for this validator's lambdas.
                            r'Is\[.*\blen\(text\)\s*>\s*30\b.*\]',
                            r"Is\[.*\btext and text\[-1\]\s*==\s*'\.'.*\]",
                            # Embeds the name of this validator's named
                            # function.
                            r'Is\[.*\b_is_quoted\b.*\]',
                        ),
                    ),
                    # String constant violating only the first of these
                    # validators.
                    HintPithUnsatisfiedMetadata(
                        pith='"Conduct my friar’s wheel"...',
                        # Match that the exception message raised documents the
                        # exact validator violated by this string.
                        exception_str_match_regexes=(
                            r'\bviolates\b.*\bIs\[.*\blen\(text\)\s*>\s*30\b.*\]',)
                    ),
                ),
            ),

            # Annotated of an isinstanceable type annotated by one beartype
            # validator synthesized from all possible operators.
            HintPepMetadata(
                hint=Annotated[str, IsLengthyOrUnquotedSentence],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying these validators.
                    HintPithSatisfiedMetadata(
                        'Dialectical, eclectic mind‐toys'),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(
                        pith=b'Of Cycladic impoverishment, cyclically unreeling',
                        # Match that the exception message raised...
                        exception_str_match_regexes=(
                            # Embeds the code for this validator's lambdas.
                            r'Is\[.*\blen\(text\)\s*>\s*30\b.*\]',
                            r"Is\[.*\btext and text\[-1\]\s*==\s*'\.'.*\]",
                            # Embeds the name of this validator's named
                            # function.
                            r'Is\[.*\b_is_quoted\b.*\]',
                        ),
                    ),
                    # String constant violating all of these validators.
                    HintPithUnsatisfiedMetadata(
                        pith='Stay its course, and',
                        # Match that the exception message raised documents the
                        # first validator violated by this string.
                        exception_str_match_regexes=(
                            r'\bviolates\b.*\bIs\[.*\blen\(text\)\s*>\s*30\b.*\]',)
                    ),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : is : &      }..............
            # Annotated of an isinstanceable type annotated by two or more
            # beartype validators all defined as functions, specified with
            # "&"-delimited operator syntax.
            HintPepMetadata(
                hint=Annotated[str, IsLengthy & IsSentence & IsQuoted],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying these validators.
                    HintPithSatisfiedMetadata(
                        '"Into piezo‐electrical, dun‐dappled lights" and...'),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(
                        pith=b'Joy.',
                        # Match that the exception message raised...
                        exception_str_match_regexes=(
                            # Embeds the code for this validator's lambdas.
                            r'Is\[.*\blen\(text\)\s*>\s*30\b.*\]',
                            r"Is\[.*\btext and text\[-1\]\s*==\s*'\.'.*\]",
                            # Embeds the name of this validator's named
                            # function.
                            r'Is\[.*\b_is_quoted\b.*\]',
                        ),
                    ),
                    # String constant violating only the first of these
                    # validators.
                    HintPithUnsatisfiedMetadata(
                        pith='"Conduct my friar’s wheel"...',
                        # Match that the exception message raised documents the
                        # exact validator violated by this string.
                        exception_str_match_regexes=(
                            r'\bviolates\b.*\bIs\[.*\blen\(text\)\s*>\s*30\b.*\]',)
                    ),
                ),
            ),

            # Annotated of an isinstanceable type annotated by two or more
            # beartype validators all defined as functions, specified with
            # "&"-delimited operator syntax such that the first such validator
            # short-circuits all subsequent such validators, which all
            # intentionally raise exceptions to prove they are silently ignored.
            #
            # Note this hint is *NOT* safely satisfiable. Ergo, we
            # intentionally do *NOT* validate this hint to be satisfied.
            HintPepMetadata(
                hint=Annotated[str, IsLengthy & IsExceptional],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(
                        b'Lone ignorance concentrations a-'),
                    # String constant violating the first validator.
                    HintPithUnsatisfiedMetadata('Long a'),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : is : |      }..............
            # Annotated of an isinstanceable type annotated by two or more
            # beartype validators all defined as functions, specified with
            # "|"-delimited operator syntax such that the first such validator
            # short-circuits all subsequent such validators, which all
            # intentionally raise exceptions to prove they are silently ignored.
            #
            # Note this hint is *NOT* safely unsatisfiable. Ergo, we
            # intentionally do *NOT* validate this hint to be unsatisfied.
            HintPepMetadata(
                hint=Annotated[str, IsLengthy | IsExceptional],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying the first validator.
                    HintPithSatisfiedMetadata(
                        "Longhouse-ignifluous, "
                        "superfluousness-rambling academic's"
                    ),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(b'Intra-convivial loci of'),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : is : nest   }..............
            # Annotated of an annotated of an isinstanceable type, each
            # annotated by a beartype validator defined as a function.
            HintPepMetadata(
                hint=Annotated[Annotated[str, IsLengthy], IsSentence],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # String constant satisfying these validators.
                    HintPithSatisfiedMetadata(
                        'Antisatellite‐dendroidal, Θṙbital Cemetery orbs — '
                        'of Moab.'
                    ),
                    # Byte-string constant *NOT* an instance of the expected
                    # type.
                    HintPithUnsatisfiedMetadata(b'Then, and'),
                    # String constant violating only the first of these
                    # validators.
                    HintPithUnsatisfiedMetadata('Though a...'),
                ),
            ),

            # List of annotateds of isinstanceable types annotated by one
            # beartype validator defined as a lambda function.
            HintPepMetadata(
                hint=List[AnnotatedStrIsLength],
                pep_sign=HintSignList,
                isinstanceable_type=list,
                is_pep585_builtin_subbed=List is list,
                piths_meta=(
                    # List of string constants satisfying this validator.
                    HintPithSatisfiedMetadata([
                        'An‐atomically Island‐stranded, adrift land)',
                        'That You randily are That worm‐tossed crabapple of',
                    ]),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        pith="Our Sturm‐sapped disorder's convolution of",
                        # Match that the exception message raised for this
                        # object embeds the code for this validator's lambda
                        # function.
                        exception_str_match_regexes=(
                            r'Is\[.*\blen\(text\)\s*>\s*30\b.*\]',),
                    ),
                    # List of string constants violating this validator.
                    HintPithUnsatisfiedMetadata([
                        'Volubly liable,',
                        'Drang‐rang aloofment –',
                        'ruthlessly',
                    ]),
                ),
            ),

            # List of lists of annotateds of an ignorable type hint annotated by
            # a type variable bounded by a beartype validator defined as a
            # lambda function.
            HintPepMetadata(
                hint=List[List[T_IsIntNonZero]],
                pep_sign=HintSignList,
                isinstanceable_type=list,
                is_pep585_builtin_subbed=List is list,
                typeargs_packed=(T_IsIntNonZero,),
                piths_meta=(
                    # List of lists of non-zero integer constants.
                    HintPithSatisfiedMetadata([[16, 17, 20], [21, 64, 65, 68]]),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        'The waves arose. Higher and higher still'),
                    # List of lists of non-integers and zeroes.
                    HintPithUnsatisfiedMetadata([
                        [
                            'Their fierce necks writhed',
                            "beneath the tempest's scourge",
                        ],
                        [0, 0, 0, 0],
                    ]),
                ),
            ),

            # List of lists of annotateds of an ignorable type hint annotated by
            # a type variable bounded by two beartype validators, one defined as
            # a lambda function and one not.
            HintPepMetadata(
                hint=List[List[T_IsIntNonZero]],
                pep_sign=HintSignList,
                isinstanceable_type=list,
                is_pep585_builtin_subbed=List is list,
                typeargs_packed=(T_IsIntNonZero,),
                piths_meta=(
                    # List of lists of non-zero integer constants.
                    HintPithSatisfiedMetadata([[16, 17, 20], [21, 64, 65, 68]]),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        'The waves arose. Higher and higher still'),
                    # List of lists of non-integers and zeroes.
                    HintPithUnsatisfiedMetadata([
                        [
                            'Their fierce necks writhed',
                            "beneath the tempest's scourge",
                        ],
                        [0, 0, 0, 0],
                    ]),
                ),
            ),

            # List of lists of annotateds of a union of isinstanceable types
            # annotated by a type variable bounded by a union of beartype
            # validators defined as lambda functions.
            HintPepMetadata(
                hint=List[List[T_IsNumberNonNegativeOrStrNonEmpty]],
                pep_sign=HintSignList,
                isinstanceable_type=list,
                is_pep585_builtin_subbed=List is list,
                typeargs_packed=(T_IsNumberNonNegativeOrStrNonEmpty,),
                piths_meta=(
                    # List of lists of positive number constants.
                    HintPithSatisfiedMetadata([
                        [11, 0.11], [1, 110, 1101100]]),
                    # List of lists of non-empty string constants.
                    HintPithSatisfiedMetadata([
                        ['The straining boat.', '—A whirlwind swept it on,'],
                        ['With fierce gusts and', 'precipitating force,'],
                    ]),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        'Through the white ridges of the chafèd sea.'),
                    # List of lists of negative numbers and empty strings.
                    HintPithUnsatisfiedMetadata([[-1, '', -0.4], ['', -5, '']]),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : isattr      }..............
            # Annotated of an isinstanceable type annotated by one beartype
            # attribute validator.
            HintPepMetadata(
                hint=Annotated[
                    CathecticallyEnsconceYouIn, IsAttrThisMobbedTristeOf],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Instance of this class satisfying this validator.
                    HintPithSatisfiedMetadata(BOSS_EMBOSSED_ORDERING),
                    # String constant *NOT* an instance of this class.
                    HintPithUnsatisfiedMetadata(
                        pith='An atoll nuclear broilers newly cleared of',
                        # Match that the exception message raised for this
                        # object embeds the name of the expected attribute.
                        exception_str_match_regexes=(
                            r"IsAttr\[.*'this_mobbed_triste_of',.*\]",),
                    ),
                    # Instance of this class *NOT* satisfying this validator.
                    HintPithUnsatisfiedMetadata(SORDIDLY_FLABBY_WRMCASTINGS),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : isequal     }..............
            # Annotated of an isinstanceable type annotated by one beartype
            # equality validator.
            HintPepMetadata(
                hint=Annotated[List[str], IsEqual[AMPLY_IMPISH]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Exact object subscripting this validator.
                    HintPithSatisfiedMetadata(AMPLY_IMPISH),
                    # Object *NOT* subscripting this validator but equal to
                    # this object.
                    HintPithSatisfiedMetadata(AMPLY_IMPISH[:]),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        pith='May Your coarsest, Incessantly cast‐off jobs of a',
                        # Match that the exception message raised for this
                        # object embeds a string in the expected list.
                        exception_str_match_regexes=(r"'Amply imp-ish'",),
                    ),
                    # List of integer constants *NOT* instances of the expected
                    # subtype.
                    HintPithUnsatisfiedMetadata([1, 2, 3, 6, 7, 14, 21, 42,]),
                    # List of string constants violating this validator.
                    HintPithUnsatisfiedMetadata(
                        ['Hectic,', 'receptacle‐hybernacling caste so',]),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : isinstance  }..............
            # Annotated of an isinstanceable type annotated by one beartype type
            # instance validator.
            HintPepMetadata(
                hint=Annotated[Class, ~IsInstance[SubclassSubclass]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Instance of the class subscripting this validator.
                    HintPithSatisfiedMetadata(Class()),
                    # Instance of the subclass subclassing the class
                    # subscripting this validator.
                    HintPithSatisfiedMetadata(Subclass()),
                    # Instance of the subsubclass subclassing the subclass
                    # subclassing the class subscripting this validator.
                    HintPithUnsatisfiedMetadata(SubclassSubclass()),
                    # Class subscripting this validator.
                    HintPithUnsatisfiedMetadata(Class),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        pith=(
                            'Architrave‐contravening, '
                            'indigenously chitinous tactilities) of'
                        ),
                        # Match that the exception message raised for this
                        # object embeds a classname in the expected list.
                        exception_str_match_regexes=(r'\bClass\b',),
                    ),
                ),
            ),

            # Type hint matching *ANY* sequence of strings, defined as the union
            # of a beartype validator matching any sequence of strings that is
            # *NOT* itself a string with a string. Although odd, this exercises
            # an obscure edge case in code generation.
            HintPepMetadata(
                hint=Union[SequenceNonstrOfStr, str],
                pep_sign=HintSignUnion,
                piths_meta=(
                    # String constant.
                    HintPithSatisfiedMetadata(
                        'Caught the impatient wandering of his gaze.'),
                    # Tuple of string constants.
                    HintPithSatisfiedMetadata((
                        'Swayed with the undulations of the tide.',
                        'A restless impulse urged him to embark',
                    )),
                    # Byte string constant.
                    HintPithUnsatisfiedMetadata(
                        pith=b'It had been long abandoned, for its sides',
                        # Match that the exception message raised for this
                        # object embeds a classname in the expected list.
                        exception_str_match_regexes=(r'\bbytes\b',),
                    ),
                    # Tuple of byte string constants.
                    HintPithUnsatisfiedMetadata(
                        pith=(
                            b'Gaped wide with many a rift,',
                            b'and its frail joints',
                        ),
                        # Match that the exception message raised for this
                        # object embeds a classname in the expected list.
                        exception_str_match_regexes=(r'\btuple\b',),
                    ),
                ),
            ),

            # ..............{ ANNOTATED ~ beartype : issubclass  }..............
            # Annotated of an isinstanceable type annotated by one beartype type
            # inheritance validator.
            HintPepMetadata(
                hint=Annotated[type, IsSubclass[Class]],
                pep_sign=HintSignAnnotated,
                piths_meta=(
                    # Class subscripting this validator.
                    HintPithSatisfiedMetadata(Class),
                    # Class subclassing the class subscripting this validator.
                    HintPithSatisfiedMetadata(Subclass),
                    # Class *NOT* subclassing the class subscripting this
                    # validator.
                    HintPithUnsatisfiedMetadata(str),
                    # String constant *NOT* an instance of the expected type.
                    HintPithUnsatisfiedMetadata(
                        pith='May Your coarsest, Incessantly cast‐off jobs of a',
                        # Match that the exception message raised for this
                        # object embeds a classname in the expected list.
                        exception_str_match_regexes=(r'\bClass\b',),
                    ),
                ),
            ),
        ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta


def hints_pep593_ignorable_deep() -> list:
    '''
    List of :pep:`593`-compliant **deeply ignorable type hints** (i.e.,
    ignorable only on the non-trivial basis of their nested child type hints).
    '''

    # ..................{ IMPORTS                            }..................
    from beartype.typing import (
        Any,
        List,
        NewType,
        Optional,
        Union,
    )
    from beartype._util.api.standard.utiltyping import get_typing_attrs

    # ..................{ LOCALS                             }..................
    # List of all PEP-specific deeply ignorable type hints to be returned.
    hints_pep_ignorable_deep = []

    # ..................{ LISTS                              }..................
    # For each PEP-specific type hint factory importable from each currently
    # importable "typing" module...
    for Annotated in get_typing_attrs('Annotated'):
        # ................{ SETS                               }................
        # Add PEP 593-specific deeply ignorable test type hints to that global.
        hints_pep_ignorable_deep.extend((
            # Annotated of shallowly ignorable type hints.
            Annotated[Any, int],
            Annotated[object, int],

            # Annotated of ignorable unions and optionals.
            Annotated[Union[Any, float, str,], int],
            Annotated[Optional[Any], int],

            # Unions and optionals of ignorable annotateds.
            Union[complex, int, Annotated[Any, int]],
            Optional[Annotated[object, int]],

            # Deeply ignorable PEP 484-, 585-, and 593-compliant type hint
            # exercising numerous edge cases broken under prior releases.
            Union[str, List[int], NewType('MetaType', Annotated[object, 53])],
        ))

    # ..................{ RETURN                             }..................
    # Return this list.
    return hints_pep_ignorable_deep



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import path hook trie** (i.e., data structure caching package names
on behalf of the higher-level :func:`beartype.claw._clawmain` submodule, which
beartype import path hooks internally created by that submodule subsequently
lookup when deciding whether or not (and how) to decorate by
:func:`beartype.beartype` the currently imported user-specific submodule).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._importlib.clawimpmain import remove_beartype_pathhook
from beartype.roar import BeartypeClawHookException
from beartype.typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatyping import CollectionStrs
from collections.abc import (
    Collection as CollectionABC,
)

# ....................{ HINTS                              }....................
PackageBasenameToTrieBlacklist = Dict[str, 'PackagesTrieBlacklist']
'''
PEP-compliant type hint matching a dictionary mapping from the unqualified
basename of each subpackage of a package to *not* be runtime type-checked on the
first importation of that subpackage to either:

* If that subpackage has *not* itself been blacklisted but merely contains one
  or more sub-subpackages that have been blacklisted, a **subpackage trie
  blacklist** (i.e., :class:`.PackagesTrieBlacklist` object recursively
  describing those sub-subpackages).
* Else, that subpackage has itself been blacklisted. In this case, the
  :data:`.PackagesTrieBlacklisted` singleton. Although *any* singleton (e.g.,
  :data:`None`, :data:`True`) would suffice here,
  :data:`.PackagesTrieBlacklisted` is the simplest and most readable.
'''


PackageBasenameToTrieWhitelist = Dict[str, 'PackagesTrieWhitelist']
'''
PEP-compliant type hint matching a dictionary mapping from the unqualified
basename of each subpackage of a package to be runtime type-checked on the first
importation of that subpackage to the **subpackage trie whitelist** (i.e.,
:class:`.PackagesTrieWhitelist` object recursively describing the
sub-subpackages of that subpackage).
'''

# ....................{ SUBCLASSES ~ blacklist             }....................
#FIXME: Unit test us up, please.
#FIXME: [SAFETY] Consider overriding the __setitem__() dunder method to ensure:
#* The passed key is a non-empty string.
#* The passed value is either "None" or another "PackagesTrieWhitelist" object.
#
#See the __init__() dunder method for similar validation logic, please.
class PackagesTrieBlacklist(PackageBasenameToTrieBlacklist):
    '''
    **(Sub)package (sub)trie blacklist** (i.e., recursively nested dictionary
    mapping from the unqualified basename of each subpackage of the current
    package to be *prevented* from being runtime type-checked on the first
    importation of that subpackage to another instance of this class similarly
    describing the sub-subpackages of that subpackage).

    This (sub)cache is suitable for caching as the values of:

    * The :data:`.packages_trie_blacklist` global dictionary.
    * Each (sub)value mapped to by that global dictionary.

    Caveats
    -------
    **This dictionary is only safely accessible in a thread-safe manner from
    within a** ``with claw_lock:`` **context manager.** Equivalently, this
    dictionary is *not* safely accessible outside that manager.

    Attributes
    ----------
    package_basename : Optional[str]
        Either:

        * If this (sub)trie is the global trie :data:`.packages_trie_blacklist`,
          :data:`None`.
        * Else, the unqualified basename of the (sub)package described by this
          (sub)trie.

    See Also
    --------
    :class:`.PackagesTrieWhitelist`
        Further details.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'package_basename',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        package_basename: Optional[str]

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Optional parameters.
        package_basename: Optional[str] = None,
        subpackage_basename_to_trie: (
            Optional[PackageBasenameToTrieBlacklist]) = None,
    ) -> None:
        '''
        Initialize this packages trie blacklist.

        Parameters
        ----------
        package_basename : Optional[str]
            Either:

            * If this is the root of a packages trie blacklist (i.e., the
              :data:`beartype.claw._clawstate.claw_state.packages_trie_blacklist`
              global variable), :data:`None`.
            * Else, the unqualified basename of the (sub)package configured by
              this (sub)trie.

            Defaults to :data:`None`.
        subpackage_basename_to_trie: Optional[PackageBasenameToTrieBlacklist]
            Either:

            * If this package trie is initially empty, :data:`None`.
            * Else, a dictionary mapping from the unqualified basename of each
              initial subpackage of this package to another **packages trie
              blacklist** (i.e,. :class:`.PackagesTrieBlacklist` class)
              describing the sub-subpackages of that subpackage. In this case,
              this packages trie blacklist is initialized from the key-value
              pairs of this dictionary.

            Defaults to :data:`None`.
        '''
        assert isinstance(package_basename, NoneTypeOr[str]), (
            f'{repr(package_basename)} neither string nor "None".')
        assert isinstance(subpackage_basename_to_trie, NoneTypeOr[dict]), (
            f'{repr(subpackage_basename_to_trie)} '
            f'neither dictionary nor "None".'
        )

        # Initialize our superclass to the empty dictionary.
        super().__init__()

        # Classify all remaining passed parameters.
        self.package_basename = package_basename

        # If the caller explicitly passed an initial dictionary to initialize
        # this dictionary subclass with...
        if subpackage_basename_to_trie:
            # If this initial dictionary is *NOT* a dictionary such that...
            if not (
                isinstance(subpackage_basename_to_trie, dict) and
                all(
                    (
                        # All keys of this dictionary are strings *AND*...
                        isinstance(subpackage_basename, str) and
                        # All values of this dictionary are either "None" or
                        # nested subpackages tries.
                        isinstance(
                            subsubpackages_trie, PackagesTrieBlacklist)
                    )
                    for subpackage_basename, subsubpackages_trie in (
                        subpackage_basename_to_trie.items())
                )
            ):
                # Raise us up the exception bomb.
                raise BeartypeClawHookException(
                    f'{repr(subpackage_basename_to_trie)} neither "None" nor '
                    f'dictionary mapping keys to packages trie blacklists.'
                )
            # Else, this initial dictionary is valid.

            # Update this dictionary subclass from this initial dictionary.
            self.update(subpackage_basename_to_trie)
        # Else, the caller explicitly passed *NO* initial dictionary. In this
        # case, preserve this dictionary subclass as the empty dictionary.

    # ..................{ COPIERS                            }..................
    #FIXME: Unit test us up, please.
    #FIXME: Comment out all of this for the moment, please. This method isn't
    #implemented correctly at the moment, sadly. It needs to call itself
    #recursively. It currently doesn't. Thus, we all sigh. *sigh*
    # def copy_deep(self) -> 'PackagesTrieBlacklist':
    #     '''
    #     Deep copy of this packages trie blacklist.
    #     '''
    #
    #     # Create and return a new deep copy of this packages trie blacklist.
    #     return PackagesTrieBlacklist(
    #         package_basename=self.package_basename,
    #         subpackage_basename_to_trie=self,
    #     )

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this packages trie blacklist.
        '''

        # If this blacklist is actually the blacklisted singleton, return a
        # truncated representation indicating this.
        if self is PackagesTrieBlacklisted:
            return 'PackagesTrieBlacklisted()'
        # Else, this blacklist is *NOT* the blacklisted singleton.

        # Return a full representation of this blacklist.
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    package_basename={repr(self.package_basename)},',
            f'    dict={super().__repr__()},',
            f')',
        ))


PackagesTrieBlacklisted = PackagesTrieBlacklist()
'''
**Blacklisted (sub)package (sub)trie** (i.e., :class:`.PackagesTrieBlacklist`
singleton arbitrarily signifying the current leaf node of a packages trie
blacklist to blacklist the corresponding (sub)package).
'''

# ....................{ SUBCLASSES ~ whitelist             }....................
class PackagesTrieWhitelist(PackageBasenameToTrieWhitelist):
    '''
    **(Sub)package (sub)trie whitelist** (i.e., recursively nested dictionary
    mapping from the unqualified basename of each subpackage of the current
    package to be runtime type-checked on the first importation of that
    subpackage to another instance of this class similarly describing the
    sub-subpackages of that subpackage).

    This (sub)cache is suitable for caching as the values of:

    * The :data:`.packages_trie_whitelist` global dictionary.
    * Each (sub)value mapped to by that global dictionary.

    Motivation
    ----------
    This dictionary is intentionally implemented as a nested trie data structure
    rather than a trivial non-nested flat dictionary. Why? Efficiency. Consider
    this flattened set of package names:

    .. code-block:: python

       package_names = {'a.b', 'a.c', 'd'}

    Deciding whether an arbitrary package name is in this set requires
    worst-case :math:`O(n)` iteration across the set of :math:`n` package names.

    Consider instead this nested trie whose keys are package names split on
    ``"."`` delimiters and whose values are either recursively nested
    dictionaries of the same format *or* the :data:`None` singleton (terminating
    the current package name):

    .. code-block:: python

       package_names_trie = {'a': {'b': None, 'c': None}, 'd': None}

    Deciding whether an arbitrary package name is in this trie only requires
    worst-case :math:`O(h)` iteration across the height :math:`h` of this
    dictionary (equivalent to the largest number of ``"."`` delimiters for any
    fully-qualified package name encapsulated by this trie). ``h <<<< n``, so
    this trie offers *much* faster worst-case lookup than that set.

    Moreover, in the worst case:

    * That set requires one inefficient string prefix test for each item.
    * This trie requires *only* one efficient string equality test for each
      nested key-value pair while descending towards the target package name.

    Let's do this, fam.

    Caveats
    -------
    **This dictionary is only safely accessible in a thread-safe manner from
    within a** ``with claw_lock:`` **context manager.** Equivalently, this
    dictionary is *not* safely accessible outside that manager.

    Attributes
    ----------
    conf_if_hooked : Optional[BeartypeConf]
        Either:

        * If this (sub)package has been explicitly registered by a prior call to
          the :func:`beartype.claw._package.clawpkgmain.hook_packages` function,
          the beartype configuration encapsulating all settings configuring
          type-checking for this (sub)package.
        * Else, :data:`None`.
    package_basename : Optional[str]
        Either:

        * If this (sub)trie is the global trie :data:`.packages_trie_whitelist`,
          :data:`None`.
        * Else, the unqualified basename of the (sub)package configured by this
          (sub)trie.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'conf_if_hooked',
        'package_basename',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        conf_if_hooked: Optional[BeartypeConf]
        package_basename: Optional[str]

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Optional parameters.
        conf_if_hooked: Optional[BeartypeConf] = None,
        package_basename: Optional[str] = None,
    ) -> None:
        '''
        Initialize this packages trie whitelist.

        Parameters
        ----------
        conf_if_hooked : Optional[BeartypeConf], defaults: None
            Either:

            * If this (sub)package has been explicitly registered by a prior
              call to the
              :func:`beartype.claw._package.clawpkgmain.hook_packages` function,
              the beartype configuration encapsulating all settings configuring
              type-checking for this (sub)package.
            * Else, :data:`None`.

            Defaults to :data:`None`.
        package_basename : Optional[str], defaults: None
            Either:

            * If this is the root of a packages trie whitelist (i.e., the
              :data:`beartype.claw._clawstate.claw_state.packages_trie_whitelist`
              global variable), :data:`None`.
            * Else, the unqualified basename of the (sub)package configured by
              this (sub)trie.

            Defaults to :data:`None`.
        '''
        assert isinstance(conf_if_hooked, NoneTypeOr[BeartypeConf]), (
            f'{repr(conf_if_hooked)} neither beartype configuration nor "None".'
        )
        assert isinstance(package_basename, NoneTypeOr[str]), (
            f'{repr(package_basename)} neither string nor "None".')

        # Initialize our superclass to the empty dictionary.
        super().__init__()

        # Classify all remaining passed parameters.
        self.conf_if_hooked = conf_if_hooked
        self.package_basename = package_basename

    # ..................{ COPIERS                            }..................
    #FIXME: Unit test us up, please.
    #FIXME: Comment out all of this for the moment, please. This method isn't
    #implemented correctly at the moment, sadly. It needs to call itself
    #recursively. It currently doesn't. Thus, we all sigh. *sigh*
    # def copy_deep(self) -> 'PackagesTrieWhitelist':
    #     '''
    #     Deep copy of this packages trie whitelist.
    #     '''
    #
    #     # Create and return a new deep copy of this packages trie whitelist.
    #     return PackagesTrieWhitelist(
    #         conf_if_hooked=self.conf_if_hooked,
    #         package_basename=self.package_basename,
    #     )

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this packages trie whitelist.
        '''

        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    package_basename={repr(self.package_basename)},',
            f'    conf_if_hooked={repr(self.conf_if_hooked)},',
            f'    dict={super().__repr__()},',
            f')',
        ))

# ....................{ RAISERS                            }....................
#FIXME: Excise us up, please. This is no longer required anywhere. *sigh*
# def die_if_packages_trie() -> None:
#     '''
#     Raise an exception if one or more packages have been registered by a prior
#     call to the :func:`beartype.claw._package.clawpkgmain.hook_packages`
#     function.
#
#     This raiser is thread-safe.
#
#     Raises
#     ------
#     BeartypeClawHookException
#         If one or more packages have been registered by a prior call to the
#         :func:`beartype.claw._package.clawpkgmain.hook_packages` function.
#     '''
#
#     # Avoid circular import dependencies.
#     from beartype.claw._clawstate import (
#         claw_lock,
#         claw_state,
#     )
#
#     # With a submodule-specific thread-safe reentrant lock...
#     with claw_lock:
#         # If one or more packages have been registered...
#         if is_packages_trie():
#             # Package trie whitelist, localized merely for readability. *sigh*
#             packages_trie_whitelist = claw_state.packages_trie_whitelist
#
#             # If a global configuration was already added by a prior call to the
#             # public beartype.claw.beartype_all() function, raise an exception.
#             if packages_trie_whitelist.conf_if_hooked is not None:
#                 raise BeartypeClawHookException(
#                     f'Prior call to package-agnostic import hook '
#                     f'beartype.claw.beartype_all() already registered '
#                     f'all packages for type-checking under '
#                     f'global beartype configuration '
#                     f'{repr(packages_trie_whitelist.conf_if_hooked)}.'
#                 )
#             # Else, or more package-specific configurations have been added by prior
#             # calls to public beartype.claw.beartype_*() functions. In this case,
#             # raise another exception.
#             else:
#                 raise BeartypeClawHookException(
#                     f'Prior call to package-specific import hook '
#                     f'beartype.claw.beartype_package() and/or '
#                     f'beartype_packages() already registered '
#                     f'{len(packages_trie_whitelist)} package(s) for '
#                     f'type-checking under beartype configurations:\n'
#                     f'{repr(packages_trie_whitelist)}'
#                 )

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_packages_trie() -> bool:
    '''
    :data:`True` only if one or more packages have been registered by a prior
    call to the :func:`beartype.claw._package.clawpkgmain.hook_packages`
    function.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Returns
    -------
    bool
        :data:`True` only if one or more packages have been registered.
    '''

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # Return true only if either...
    return (
        # A global configuration has been added by a prior call to the
        # beartype.claw.beartype_all() import hook *OR*...
        claw_state.packages_trie_whitelist.conf_if_hooked is not None or
        # One or more package-specific configurations have been added by
        # prior calls to either the beartype.claw.beartype_package() *OR*
        # beartype_packages() import hooks.
        bool(claw_state.packages_trie_whitelist)
    )


#FIXME: Unit test us up, please.
def is_package_blacklisted(package_basenames: CollectionStrs) -> bool:
    '''
    :data:`True` only if the package with the passed name has been
    **blacklisted** (i.e., prevented from being runtime type-checked on the
    first importation of that package) by being either:

    * Explicitly blacklisted by being directly listed in a previously configured
      :attr:`beartype.BeartypeConf.claw_skip_package_names` collection.
    * Implicitly blacklisted by being the subpackage of a parent package
      directly listed in such a collection.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Parameters
    ----------
    package_basenames : CollectionStrs
        Collection of each unqualified basename comprising the fully-qualified
        name of the package to be inspected, split by the caller from this
        name on ``"."`` delimiters.

    Returns
    -------
    bool
        :data:`True` only if this package has been blacklisted.
    '''
    assert isinstance(package_basenames, CollectionABC), (
        f'{repr(package_basenames)} not collection.')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # ....................{ LOCALS                         }....................
    # True only if this package has been blacklisted.
    is_blacklisted = False

    # Current subtrie of the global trie blacklist describing the currently
    # iterated basename of each parent package of this package to be blacklisted
    # (i.e., ignored), initialized to this global trie.
    subpackages_trie_blacklist: Optional[PackagesTrieBlacklist] = (
        claw_state.packages_trie_blacklist)

    # ....................{ SEARCH                         }....................
    # For each unqualified basename of each parent package transitively
    # containing this package (as well as that of this package itself)...
    for package_basename in package_basenames:
        # print(f'Visiting blacklisting parent package "{package_basename}"...')

        # Current subtrie of this trie blacklist describing this parent package
        # if this parent package contains one or more subpackages that have been
        # blacklisted by a prior configuration of the
        # "BeartypeConf.claw_skip_package_names" list *OR* "None" otherwise
        # (i.e., if this parent package has yet to be blacklisted).
        subpackages_trie_blacklist = subpackages_trie_blacklist.get(  # type: ignore[union-attr]
            package_basename)

        # If *NO* subpackages of this parent package have been blacklisted,
        # halt iteration.
        if subpackages_trie_blacklist is None:
            break
        # Else, one or more subpackages of this parent package have been
        # blacklisted.
        #
        # If this parent package contains *NO* subpackages, this is a leaf
        # (i.e., terminal) subtrie. In this case, the subpackage of this parent
        # package that has been blacklisted is this parent package itself.
        # Return true immediately.
        #
        # You are now thinking: "B-b-but how can a package be a subpackage of
        # itself?" Simple. In the same set theoretic sense that all classes are
        # subclasses of themselves and all sets are subsets of themselves, all
        # packages are subpackages of themselves. \o/
        elif subpackages_trie_blacklist is PackagesTrieBlacklisted:
            # print(f'Skipping blacklisted package "{package_basename}"...')
            is_blacklisted = True
            break
        # Else, this parent package contains one or more subpackages. In this
        # case, continue iterating until exhausting all subtries *OR* visiting a
        # leaf subtrie.
    # Else, neither this package *NOR* a parent package of this package has
    # been blacklisted. In this case, this package *COULD* still have been
    # whitelisted. Proceed to the next phase, Dr. Demento!

    # Return this boolean.
    return is_blacklisted

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_package_conf_or_none(package_name: str) -> Optional[BeartypeConf]:
    '''
    Beartype configuration with which to type-check the package with the passed
    name if that package *or* a parent package of that package was registered by
    a prior call to the :func:`.hook_packages` function *or* :data:`None`
    otherwise (i.e., if neither that package *nor* a parent package of that
    package was registered by such a call).

    This getter is thread-safe.

    Parameters
    ----------
    package_name : str
        Fully-qualified name of the package to be inspected.

    Returns
    -------
    Optional[BeartypeConf]
        Either:

        * If that package or a parent package of that package was registered by
          a prior call to the :func:`.hook_packages` function, the beartype
          configuration with which to type-check that package.
        * Else, :data:`None`.
    '''

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.claw._clawstate import (
        claw_lock,
        claw_state,
    )

    # ....................{ LOCALS                         }....................
    # Beartype configuration to be returned, defaulting to "None".
    subpackage_conf: Optional[BeartypeConf] = None

    # List of each unqualified basename comprising this name, split from this
    # fully-qualified name on "." delimiters. Note that the "str.split('.')" and
    # "str.rsplit('.')" calls produce the exact same lists under all possible
    # edge cases. We arbitrarily call the former rather than the latter for
    # simplicity and readability.
    package_basenames = package_name.split('.')

    # ....................{ SEARCH                         }....................
    # With a submodule-specific thread-safe reentrant lock...
    with claw_lock:
        # print(f'claw_state: {claw_state}')

        # ....................{ PHASE 1 ~ blacklist        }....................
        # In this first phase, decide whether this package has been either:
        # * Explicitly blacklisted by being directly listed in a previously
        #   configured "BeartypeConf.claw_skip_package_names" collection.
        # * Implicitly blacklisted by being the subpackage of a parent package
        #   directly listed in such a collection.
        #
        # If either of these is the case, this getter function *IMMEDIATELY*
        # reduces to a noop by returning "None". For that reason, these two
        # phases *CANNOT* be efficiently interleaved with one another. Before
        # the second phase returns *ANYTHING*, the first phase decides whether
        # the second phase should even be performed at all.
        #
        # If this package has *NOT* been blacklisted...
        if not is_package_blacklisted(package_basenames):
            # ....................{ PHASE 2 ~ whitelist    }....................
            # In this second phase, decide whether this package has been either:
            # * Explicitly whitelisted by being directly passed to a public
            #   "beartype.claw" import hook (e.g., beartype_package()).
            # * Implicitly blacklisted by being the subpackage of a parent
            #   package directly passed to such an import hook.

            # Beartype configuration registered for the currently iterated
            # package, defaulting to the beartype configuration registered for
            # the global trie applicable to *ALL* packages if an external caller
            # previously called the public beartype.claw.beartype_all() function
            # *OR* "None" otherwise (i.e., if that function has yet to be
            # called).
            subpackage_conf = claw_state.packages_trie_whitelist.conf_if_hooked

            # For each subpackages trie describing each parent package
            # transitively containing this package (as well as that of that
            # package itself)...
            for subpackages_trie in iter_packages_trie(package_basenames):
                # Beartype configuration registered with either...
                subpackage_conf = (
                    # That parent package if any *OR*...
                    #
                    # Since that parent package is more granular (i.e., unique)
                    # than any transitive parent package of that parent package,
                    # the former takes precedence over the latter when defined.
                    subpackages_trie.conf_if_hooked or
                    # A transitive parent package of that parent package if any.
                    subpackage_conf
                )
            # print(f'Discovered package "{package_name}" beartype conf {repr(subpackage_conf)}!')
        # Else, this package has been blacklisted.
        # else:
        #     print(f'Skipping blacklisted package "{package_name}"...')

    # ....................{ RETURN                         }....................
    # Return this beartype configuration if any *OR* "None" otherwise.
    return subpackage_conf

# ....................{ ITERATORS                          }....................
#FIXME: Unit test us up, please.
def iter_packages_trie(
    package_basenames: CollectionStrs) -> Iterable[PackagesTrieWhitelist]:
    '''
    Thread-safe generator iteratively yielding one **(sub)package (sub)trie
    whitelist** (i.e., :class:`PackagesTrieWhitelist` instance) describing each
    transitive parent package of the package with the passed name if this
    package or a parent package of this package was hooked by a prior call to
    the :func:`beartype.claw._package.clawpkgmain.hook_packages` function *or*
    the empty iterable otherwise otherwise (i.e., if neither this package nor a
    parent package of this package was hooked by such a call).

    Specifically, this generator yields (in order):

    #. The subtrie of this trie configuring the root package of the passed
       (sub)package.
    #. And so on, until eventually yielding...
    #. The subsubtrie of this subtrie configuring the passed (sub)package
       itself.

    This generator intentionally avoids yielding the global trie
    :data:`beartype.claw._clawstate.packages_trie_whitelist`, which is already
    accessible via that global.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Parameters
    ----------
    package_basenames : CollectionStrs
        Collection of each unqualified basename comprising the fully-qualified
        name of the package to be inspected, split by the caller from this
        name on ``"."`` delimiters.

    Yields
    ------
    PackagesTrieWhitelist
        (Sub)package configuration (sub)trie describing the currently iterated
        transitive parent package of the package with this name.
    '''
    assert isinstance(package_basenames, CollectionABC), (
        f'{repr(package_basenames)} not collection.')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # ....................{ LOCALS                         }....................
    # Current subtrie of the global trie whitelist describing the currently
    # iterated basename of each parent package of this package to be
    # whitelisted (i.e., hooked), initialized to this global trie.
    subpackages_trie_whitelist: Optional[PackagesTrieWhitelist] = (  # type: ignore[union-attr]
        claw_state.packages_trie_whitelist)

    # ....................{ SEARCH                         }....................
    # For each unqualified basename of each parent package transitively
    # containing this package (as well as that of this package itself)...
    for package_basename in package_basenames:
        # Current subtrie of this trie whitelist describing this parent package
        # if this parent package was hooked by a prior call to the
        # hook_packages() function *OR* "None" otherwise (i.e., if this parent
        # package has yet to be hooked).
        subpackages_trie_whitelist = subpackages_trie_whitelist.get(  # type: ignore[union-attr]
            package_basename)

        # If this parent package has yet to be hooked, halt iteration.
        if subpackages_trie_whitelist is None:
            break
        # Else, this parent package was previously hooked.

        # Yield this subtrie whitelist describing this parent package.
        yield subpackages_trie_whitelist

# ....................{ REMOVERS                           }....................
#FIXME: Unit test us up, please.
def remove_beartype_pathhook_unless_packages_trie() -> None:
    '''
    Remove our **beartype import path hook singleton** (i.e., single callable
    guaranteed to be inserted at most once to the front of the standard
    :mod:`sys.path_hooks` list recursively applying the
    :func:`beartype.beartype` decorator to all well-typed callables and classes
    defined by all submodules of all packages previously registered by a call to
    a public :func:`beartype.claw` function) if this path hook has already been
    added and all previously registered packages have been unregistered *or*
    silently reduce to a noop otherwise (i.e., if either this path hook has yet
    to be added or one or more packages are still registered).

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.
    '''

    # If all previously registered packages have been unregistered, safely
    # remove our import path hook from the "sys.path_hooks" list.
    if not is_packages_trie():
        remove_beartype_pathhook()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Semantically empty submodule containing only only zero or more whitespace
characters zero or more inline comments, and this prefacing module docstring --
exercising a pernicious edge case in AST transformation that actually happened
and destroyed everything.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **hint overrides class hierarchy** (i.e., public classes implementing
immutable mappings intended to be passed as the value of the ``hint_overrides``
parameter accepted by the :class:`beartype.BeartypeConf.__init__` method).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeConfParamException
from beartype.typing import Optional
from beartype._data.typing.datatyping import (
    DictStrToAny,
    Pep484TowerComplex,
    Pep484TowerFloat,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict

# ....................{ GETTERS                            }....................
def sanify_conf_kwargs_is_pep484_tower(conf_kwargs: DictStrToAny) -> None:
    '''
    Sanify (i.e., sanitize) the passed ``is_pep484_tower`` property of the
    passed dictionary of configuration parameters.

    Parameters
    ----------
    conf_kwargs : Dict[str, object]
        Dictionary mapping from the names to values of *all* possible keyword
        parameters configuring this configuration.
    '''
    assert isinstance(conf_kwargs, dict), f'{repr(conf_kwargs)} not dictionary.'

    # ....................{ LOCALS                         }....................
    # PEP 484-compliant implicit tower type hint overrides (i.e., "FrozenDict"
    # instance lossily convering integers to floating-point numbers *AND* both
    # integers and floating-point numbers to complex numbers).
    HINT_OVERRIDES_PEP484_TOWER = _hint_overrides_pep484_tower()

    # Hint overrides if passed by the caller *OR* "None" otherwise.
    hint_overrides = conf_kwargs['hint_overrides']

    # Target hint overrides for the source "float" and "complex" types if any
    # *OR* "None" otherwise.
    hint_overrides_float = hint_overrides.get(float)
    hint_overrides_complex = hint_overrides.get(complex)

    # Whichever of the "float" or "complex" types are already existing overrides
    # in the passed type hint overrides.
    hint_override_cls_conflict: Optional[type] = None

    # ....................{ TEST                           }....................
    # If these overrides already define conflicting overrides for either the
    # "float" or "complex" types, record that fact.
    if (
        hint_overrides_float and
        hint_overrides_float != HINT_OVERRIDES_PEP484_TOWER[float]
    ):
        hint_override_cls_conflict = float
    elif (
        hint_overrides_complex and
        hint_overrides_complex != HINT_OVERRIDES_PEP484_TOWER[complex]
    ):
        hint_override_cls_conflict = complex
    # Else, these overrides do *NOT* already define conflicting overrides
    # for either the "float" or "complex" types.

    # If these overrides already define conflicting overrides for either the
    # "float" or "complex" types, raise an exception.
    if hint_override_cls_conflict:
        raise BeartypeConfParamException(
            f'Beartype configuration '
            f'parameter "is_pep484_tower" conflicts with '
            f'parameter "hint_overrides" key '
            f'"{hint_override_cls_conflict.__name__}" value '
            f'{repr(hint_overrides[hint_override_cls_conflict])}.'
        )
    # Else, these overrides do *NOT* already define conflicting overrides
    # for either the "float" or "complex" types.

    # ....................{ SANIFY                         }....................
    # Add hint overrides expanding the passed type hint overrides with
    # additional overrides mapping:
    # * The "float" type to the "float | int" type hint.
    # * The "complex" type to the "complex | float | int" type hint.
    conf_kwargs['hint_overrides'] = hint_overrides | HINT_OVERRIDES_PEP484_TOWER  # type: ignore[assignment]

# ....................{ FACTORIES                          }....................
@callable_cached
def _hint_overrides_pep484_tower() -> FrozenDict:
    '''
    :pep:`484`-compliant **implicit tower type hint overrides** (i.e.,
    :class:`.FrozenDict` instance lossily converting integers to floating-point
    numbers *and* both integers and floating-point numbers to complex numbers).

    Specifically, these overrides instruct :mod:`beartype` to automatically
    expand:

    * All :class:`float` type hints to ``float | int``, thus implicitly
      accepting both integers and floating-point numbers for objects annotated
      as only accepting floating-point numbers.
    * All :class:`complex` type hints to ``complex | float | int``, thus
      implicitly accepting integers, floating-point, and complex numbers for
      objects annotated as only accepting complex numbers.

    This getter is memoized for efficiency. Note that this getter is
    intentionally defined as a memoized function rather than a global variable
    of this submodule. Why? Because the latter approach induces a circular
    import dependency. (I sigh.)
    '''

    # Beartype on the job, Sir!
    return FrozenDict({float: Pep484TowerFloat, complex: Pep484TowerComplex,})



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`749`-compliant **deferred type hints** (i.e., hints that are
possibly unquoted forward references referring to currently undefined types
under Python >= 3.14).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep749Exception
from beartype._cave._cavefast import Format  # pyright: ignore
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatyping import TypeException
from beartype._data.typing.datatypingport import (
    Hint,
    HintOrSentinel,
)
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
#FIXME: Revise docstring, please. *sigh*
def get_hint_pep749_subhint_mandatory(
    # Mandatory parameters.
    hint: Hint,
    subhint_name_dynamic: str,
    subhint_name_static: str,

    # Optional parameters.
    hint_format: Format = Format.FORWARDREF,
    exception_cls: TypeException = BeartypeDecorHintPep749Exception,
    exception_prefix: str = '',
) -> Hint:
    '''
    :pep:`749`-compliant **mandatory subhint** (i.e., nested type hint defined
    as an attribute on a parent type hint whose value is required to *not* be
    :data:`None` but may be a possibly unquoted forward reference referring to a
    currently undefined type) with the passed static or dynamic attribute names
    defined on the passed parent type hint if any *or* raise an exception
    otherwise (i.e., if this parent type hint either fails to define an
    attribute having either of these names or defines such an attribute to be
    :data:`None`).

    This getter is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces to
    a one-liner.

    Caveats
    -------
    **This getter is only intended to be passed** :pep:`749`-compliant **hints
    whose subhints are mandatory** (rather than optional). Currently, the only
    mandatory subhints are:

    * :pep:`695`-compliant ``type`` alias value subhints defined as the
      ``__value__`` dunder attribute.

    Parameters
    ----------
    hint : Hint
        Parent type hint to be inspected.
    subhint_name_dynamic : str
        Unqualified basename of the attribute defined on this parent type hint
        yielding this hint's **dynamic subhint** (i.e., bound method accepting
        the passed format and returning this subhint in this format).
    subhint_name_static : str
        Unqualified basename of the attribute defined on this parent type hint
        yielding this hint's **static subhint** (i.e., dunder instance variable
        whose value directly provides this subhint).
    hint_format : Format, default: Format.FORWARDREF
        Format of annotated hints to be returned. Defaults to
        :attr:`Format.FORWARDREF`, in which case this getter safely encapsulates
        each otherwise unsafe unquoted forward reference transitively
        subscripting each hint annotating this hintable with a safe
        :class:`annotationlib.ForwardRef` object. Note that the remaining
        formats are situational at best. See also the
        :func`beartype._util.hint.pep.proposal.pep649.get_pep649_hintable_annotations`
        getter for further details.
    exception_cls : TypeException, default: BeartypeDecorHintPep749Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep749Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Hint
        Subhint with this static or dynamic attribute name defined on this hint.

    Raises
    ------
    exception_cls
        If this parent type hint either:

        * Fails to define an attribute having either of these names.
        * Defines an attribute having either of these names to be **optional**
          (i.e., whose value is :data:`None`).
    '''

    # Subhint defined by this hintable if any *OR* the sentinel placeholder.
    hint_subhint = get_hint_pep749_subhint_optional(
        hint=hint,
        subhint_name_dynamic=subhint_name_dynamic,
        subhint_name_static=subhint_name_static,
        # Since a mandatory subhint by definition has *NO* default, select an
        # arbitrary beartype-specific default guaranteed to *NEVER* be a valid
        # type hint. The sentinel placeholder wins again.
        subhint_value_null=SENTINEL,
        hint_format=hint_format,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If this hint nullifies this subhint, raise an exception.
    #
    # Note that this should *NEVER* occur and cannot, in fact, be tested. *sigh*
    if hint_subhint is SENTINEL:  # pragma: no cover
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not class.')
        assert issubclass(exception_cls, Exception), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise a human-readable exception.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'mandatory subhint nullified (i.e., '
            f'static subhint attribute "{subhint_name_static}" and '
            f'dynamic subhint attribute "{subhint_name_dynamic}" '
            f'both defined to be "None").'
        )
    # Else, this hint does *NOT* nullify this subhint.

    # Return this subhint.
    return hint_subhint  # pyright: ignore

# ....................{ PRIVATE ~ getters                  }....................
def _get_hint_pep749_subhint_optional_static(  # pyright: ignore
    # Mandatory parameters.
    hint: Hint,
    subhint_name_static: str,
    subhint_value_null: object,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep749Exception,
    exception_prefix: str = '',
) -> HintOrSentinel:
    '''
    :pep:`749`-compliant **optional static subhint** (i.e., nested type hint
    defined as an attribute on a parent type hint whose value may be either the
    passed null subhint value or a possibly unquoted forward reference referring
    to a currently undefined type) with the passed static attribute name defined
    on the passed parent type hint if this value is not this null subhint value
    *or* the sentinel placeholder otherwise (i.e., if this parent type hint
    defines an attribute having this name whose value is this null subhint
    value).

    This getter is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces to
    a one-liner.

    Parameters
    ----------
    hint : Hint
        Parent type hint to be inspected.
    subhint_name_static : str
        Unqualified basename of the attribute defined on this parent type hint
        yielding this hint's **static subhint** (i.e., dunder instance variable
        whose value directly provides this subhint).
    subhint_value_null : object
        **Null subhint value** (i.e., arbitrary object signifying this subhint
        to be nullified), implying this subhint to be unspecified.
    exception_cls : TypeException, default: BeartypeDecorHintPep749Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep749Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Hint
        If this parent type hint defines an attribute having this name whose
        value is either:

        * The passed null subhint value, the sentinel placeholder.
        * Else, the value of that attribute.

    Raises
    ------
    exception_cls
        If this parent type hint fails to define an attribute having this name.
    '''
    assert isinstance(subhint_name_static, str), (
        f'{repr(subhint_name_static)} not string.')

    # Subhint to be returned. Specifically, either:
    # * If this hint defines a static subhint with this name, the value of this
    #   subhint: a simple PEP-compliant child hint.
    # * Else, the sentinel placeholder.
    subhint_value = getattr(hint, subhint_name_static, SENTINEL)

    # If this hint defines *NO* static subhint with this name, raise an
    # exception.
    #
    # Note that this should *NEVER* be the case.
    if subhint_value is SENTINEL:
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'static subhint attribute "{subhint_name_static}" undefined.'
        )
    # Else, this hint defines a static subhint with this name.
    #
    # If this subhint is the null value, silently reduce to a noop and
    # immediately coerce this subhint to the sentinel placeholder.
    elif subhint_value is subhint_value_null:
        subhint_value = SENTINEL
    # Else, this subhint is *NOT* the null value. In this case, preserve this
    # subhint as is.

    # Return this subhint.
    return subhint_value

# ....................{ VERSIONS                           }....................
# If the active Python interpreter targets Python >= 3.14...
if IS_PYTHON_AT_LEAST_3_14:
    # ....................{ IMPORTS                        }....................
    # Defer version-specific imports.
    from annotationlib import call_evaluate_function  # type: ignore[import-not-found]

    # ....................{ GETTERS                        }....................
    #FIXME: Consider memoization. This getter is likely to be *EXTREMELY*
    #slow. That said, we are simply outta volunteer time on this one.
    #Sometimes, you just gotta ship it as is. *shrug*
    def get_hint_pep749_subhint_optional(  # pyright: ignore
        # Mandatory parameters.
        hint: Hint,
        subhint_name_dynamic: str,
        subhint_name_static: str,
        subhint_value_null: object,

        # Optional parameters.
        hint_format: Format = Format.FORWARDREF,
        exception_cls: TypeException = BeartypeDecorHintPep749Exception,
        exception_prefix: str = '',
    ) -> HintOrSentinel:
        assert isinstance(subhint_name_dynamic, str), (
            f'{repr(subhint_name_dynamic)} not string.')
        assert isinstance(hint_format, Format), (
            f'{repr(hint_format)} not annotation format.')

        # If this hint format is the forward reference format implicitly
        # coercing nested unquoted forward references into proxy objects...
        #
        # Note that this is the most common hint format and thus intentionally
        # tested first.
        if hint_format is Format.FORWARDREF:
            # Attempt to first efficiently retrieve this subhint's static value.
            try:
                subhint_value = _get_hint_pep749_subhint_optional_static(
                    hint=hint,
                    subhint_name_static=subhint_name_static,
                    subhint_value_null=subhint_value_null,
                    exception_cls=exception_cls,
                    exception_prefix=exception_prefix,
                )
            # If doing so raises a builtin "NameError" exception, this subhint
            # contains one or more unquoted forward references. In this case...
            except NameError:
                # Fallback to inefficiently retrieving this subhint's dynamic
                # value coercing these unquoted forward references into proxies.
                subhint_value = _get_hint_pep749_subhint_optional_dynamic(
                    hint=hint,
                    subhint_name_dynamic=subhint_name_dynamic,
                    subhint_value_null=subhint_value_null,
                    hint_format=hint_format,
                    exception_cls=exception_cls,
                    exception_prefix=exception_prefix,
                )
        # Else, this hint format is *NOT* the forward reference format.
        #
        # If this hint format is the value format raising exceptions if this
        # subhint contains one or more unquoted forward references...
        elif hint_format is Format.VALUE:
            # Trivially defer to this private static-specific getter.
            subhint_value = _get_hint_pep749_subhint_optional_static(
                hint=hint,
                subhint_name_static=subhint_name_static,
                subhint_value_null=subhint_value_null,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            )
        # Else, this hint format is *NOT* the value format. In this case...
        else:
            # Fallback to inefficiently retrieving this subhint's dynamic value
            # coercing this subhint into this format.
            subhint_value = _get_hint_pep749_subhint_optional_dynamic(
                hint=hint,
                subhint_name_dynamic=subhint_name_dynamic,
                subhint_value_null=subhint_value_null,
                hint_format=hint_format,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            )

        # Return this subhint.
        return subhint_value

    # ....................{ PRIVATE ~ getters              }....................
    def _get_hint_pep749_subhint_optional_dynamic(  # pyright: ignore
        # Mandatory parameters.
        hint: Hint,
        subhint_name_dynamic: str,
        subhint_value_null: object,

        # Optional parameters.
        hint_format: Format = Format.FORWARDREF,
        exception_cls: TypeException = BeartypeDecorHintPep749Exception,
        exception_prefix: str = '',
    ) -> HintOrSentinel:
        '''
        :pep:`749`-compliant **optional dynamic subhint** (i.e., nested type hint
        defined as an attribute on a parent type hint whose value may be either the
        passed null subhint value or a low-level C-based evaluator function
        implemented by CPython whose signature matches ``Callable[[Format], Hint]``)
        with the passed dynamic attribute name defined by calling this evaluator
        function on the passed parent type hint if this value is not this null
        subhint value *or* the sentinel placeholder otherwise (i.e., if this parent
        type hint defines an attribute having this name whose value is this null
        subhint value).

        This getter is intentionally *not* memoized (e.g., by the
        ``@callable_cached`` decorator), as the implementation trivially reduces to
        a one-liner.

        Parameters
        ----------
        hint : Hint
            Parent type hint to be inspected.
        subhint_name_dynamic : str
            Unqualified basename of the attribute defined on this parent type hint
            yielding this hint's **dynamic subhint** (i.e., bound method accepting
            the passed format and returning this subhint in this format).
        subhint_value_null : object
            **Null subhint value** (i.e., arbitrary object signifying this subhint
            to be nullified), implying this subhint to be unspecified.
        hint_format : Format, default: Format.FORWARDREF
            Format of annotated hints to be returned. Defaults to
            :attr:`Format.FORWARDREF`, in which case this getter safely encapsulates
            each otherwise unsafe unquoted forward reference transitively
            subscripting each hint annotating this hintable with a safe
            :class:`annotationlib.ForwardRef` object. Note that the remaining
            formats are situational at best. See also the
            :func`beartype._util.hint.pep.proposal.pep649.get_pep649_hintable_annotations`
            getter for further details.
        exception_cls : TypeException, default: BeartypeDecorHintPep749Exception
            Type of exception to be raised in the event of a fatal error. Defaults
            to :exc:`.BeartypeDecorHintPep749Exception`.
        exception_prefix : str, default: ''
            Human-readable substring prefixing raised exception messages. Defaults
            to the empty string.

        Returns
        -------
        Hint
            If this parent type hint defines an attribute having this name whose
            value is either:

            * The passed null subhint value, the sentinel placeholder.
            * Else, the value of that attribute.

        Raises
        ------
        exception_cls
            If this parent type hint fails to define an attribute having this name.
        '''
        assert isinstance(subhint_name_dynamic, str), (
            f'{repr(subhint_name_dynamic)} not string.')

        # Subhint to be returned. Specifically, either:
        # * If this hint defines a dynamic subhint with this name, the value of this
        #   subhint: a low-level C-based evaluator function implemented by CPython,
        #   whose signature matches "Callable[[Format], Hint]".
        # * Else, the sentinel placeholder.
        subhint_value = getattr(hint, subhint_name_dynamic, SENTINEL)

        # If this hint defines *NO* dynamic subhint with this name, raise an
        # exception. Note that this should *NEVER* be the case.
        if subhint_value is SENTINEL:
            raise exception_cls(
                f'{exception_prefix}type hint {repr(hint)} '
                f'dynamic subhint attribute "{subhint_name_dynamic}" undefined.'
            )
        # Else, this hint defines a dynamic subhint with this name.
        #
        # If this subhint is the null value, silently reduce to a noop and
        # immediately coerce this subhint to the sentinel placeholder.
        elif subhint_value is subhint_value_null:
            subhint_value = SENTINEL
        # Else, this subhint is *NOT* the null value.
        #
        # If this subhint is uncallable, this subhint is *NOT* an evaluator
        # function. In this case, raise an exception.
        #
        # Note that this should *NEVER* be the case.
        elif not callable(subhint_value):
            raise exception_cls(
                f'{repr(subhint_value)} uncallable.')
        # Else, this subhint is callable and thus an evaluator function. In
        # this case...
        else:
            # Subhint formatted by this evaluator function in this format.
            subhint_value = call_evaluate_function(subhint_value, hint_format)

            # If this subhint is the null value, silently reduce to a noop and
            # immediately coerce this subhint to the sentinel placeholder.
            #
            # Note that this redundancy intentionally exists to handle obscure (and
            # presumably unintentional edge cases) in PEP 749. Notably:
            #     >>> from typing import TypeVar
            #     >>> T = TypeVar('T')
            #     >>> T.__bound__
            #     None  # <-- makes sense
            #     >>> T.evaluate_bound
            #     None  # <-- makes sense, sorta
            #     >>> T.__default__
            #     typing.NoDefault  # <-- makes no sense, but okay
            #     >>> T.evaluate_default
            #     <constevaluator typing.NoDefault>  # <-- makes no sense and contradicts PEP 749
            #     >>> T.evaluate_default(1)
            #     typing.NoDefault  # <-- makes no sense, but okay
            if subhint_value is subhint_value_null:
                subhint_value = SENTINEL
            # Else, this subhint is *NOT* the null value.

        # Return this subhint.
        return subhint_value

# Else, the active Python interpreter targets Python <= 3.13. In this case,
# trivially defer to the PEP 484-compliant "__annotations__" dunder attribute.
else:
    # ....................{ GETTERS                        }....................
    def get_hint_pep749_subhint_optional(
        # Mandatory parameters.
        hint: Hint,
        subhint_name_dynamic: str,
        subhint_name_static: str,
        subhint_value_null: object,

        # Optional parameters.
        hint_format: Format = Format.FORWARDREF,
        exception_cls: TypeException = BeartypeDecorHintPep749Exception,
        exception_prefix: str = '',
    ) -> HintOrSentinel:

        # Trivially defer to the private static-specific getter defined above.
        # Note that this call intentionally omits irrelevant parameters. Ergo,
        # these two getters *CANNOT* simply be aliased to one another.
        return _get_hint_pep749_subhint_optional_static(
            hint=hint,
            subhint_name_static=subhint_name_static,
            subhint_value_null=subhint_value_null,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )


get_hint_pep749_subhint_optional.__doc__ = (
    '''
    :pep:`749`-compliant **optional subhint** (i.e., nested type hint defined
    as an attribute on a parent type hint whose value may be either :data:`None`
    or a possibly unquoted forward reference referring to a currently undefined
    type) with the passed static or dynamic attribute names defined on the
    passed parent type hint if this value is not :data:`None` *or* the sentinel
    placeholder otherwise (i.e., if this parent type hint defines an attribute
    having either of these names whose value is the passed null subhint value).

    This getter intentionally returns the sentinel placeholder (rather than
    :data:`None`) when either of these attributes is the null subhint value.
    Why? To enable callers to distinguish between subhints that are nullified
    from subhints whose values are actually :data:`None`, which is (of course) a
    valid :pep:`484`-compliant type hint.

    This getter is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces to
    a one-liner.

    Caveats
    -------
    **This getter is only intended to be passed** :pep:`749`-compliant **hints
    whose subhints are optional** (rather than mandatory). Currently, the only
    optional subhints are:

    * :pep:`484`-compliant type variable bound and constraint subhints defined
      as the ``__bound__`` and ``__constraints__`` dunder attributes.
    * :pep:`696`-compliant type variable default subhints defined as the
      ``__default__`` dunder attribute.

    Parameters
    ----------
    hint : Hint
        Parent type hint to be inspected.
    subhint_name_dynamic : str
        Unqualified basename of the attribute defined on this parent type hint
        yielding this hint's **dynamic subhint** (i.e., bound method accepting
        the passed format and returning this subhint in this format).
    subhint_name_static : str
        Unqualified basename of the attribute defined on this parent type hint
        yielding this hint's **static subhint** (i.e., dunder instance variable
        whose value directly provides this subhint).
    subhint_value_null : object
        **Null subhint value** (i.e., arbitrary object signifying this subhint
        to be nullified), implying this subhint to be unspecified.
    hint_format : Format, default: Format.FORWARDREF
        Format of annotated hints to be returned. Defaults to
        :attr:`Format.FORWARDREF`, in which case this getter safely encapsulates
        each otherwise unsafe unquoted forward reference transitively
        subscripting each hint annotating this hintable with a safe
        :class:`annotationlib.ForwardRef` object. Note that the remaining
        formats are situational at best. See also the
        :func`beartype._util.hint.pep.proposal.pep649.get_pep649_hintable_annotations`
        getter for further details.
    exception_cls : TypeException, default: BeartypeDecorHintPep749Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep749Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Hint
        If this parent type hint defines an attribute having either of these
        names whose value is either:

        * The passed null subhint value, the sentinel placeholder.
        * Else, the value of that attribute.

    Raises
    ------
    exception_cls
        If this parent type hint fails to define an attribute having either of
        these names.
    '''
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-compliant type hint getters** (i.e., low-level callables
introspecting PEP-compliant type hints for associated metadata).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPepException
from beartype.roar._roarexc import _BeartypeUtilTypeException
from beartype.typing import (
    Any,
    Optional,
    Union,
)
from beartype._cave._cavefast import HintPep646TypeVarTupleType
from beartype._data.cls.datacls import TYPES_NONPEP_TYPEARGS_PACKED
from beartype._data.typing.datatypingport import (
    Hint,
    HintOrNone,
)
from beartype._data.typing.datatyping import (
    TuplePep484612646TypeArgsPacked,
    TuplePep484612646TypeArgsUnpacked,
    TypeException,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_ORIGIN_ISINSTANCEABLE,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.proposal.pep585 import (
    get_hint_pep585_generic_typeargs_packed,
    is_hint_pep585_generic_unsubbed,
)

# ....................{ GETTERS ~ args                     }....................
def get_hint_pep_args(hint: object) -> tuple:
    '''
    Tuple of the zero or more **child type hints** subscripting (indexing) the
    passed PEP-compliant type hint if this hint was subscripted *or* the empty
    tuple otherwise (i.e., if this hint is unsubscripted and is thus either an
    unsubscriptable type hint *or* a subscriptable type hint factory that is
    unsubscripted).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This getter should always be called in lieu of attempting to directly
    access the low-level** ``__args__`` **dunder attribute.** Various
    singleton objects defined by the :mod:`typing` module (e.g.,
    :attr:`typing.Any`, :attr:`typing.NoReturn`) fail to define this attribute,
    guaranteeing :class:`AttributeError` exceptions from all general-purpose
    logic attempting to directly access this attribute. Thus this function,
    which "fills in the gaps" by implementing this oversight.

    **This getter lies rarely due to subscription erasure** (i.e., the malicious
    destruction of child type hints by parent type hint factories at
    subscription time). Callers should *not* assume that the objects originally
    subscripting this hint are still accessible. Although *most* hints preserve
    their subscripted objects over their lifetimes, a small subset of edge-case
    hints erase those objects at subscription time. This includes:

    * :pep:`585`-compliant empty tuple type hints (i.e., ``tuple[()]``), which
      despite being explicitly subscripted erroneously erase that subscription
      at subscription time. This does *not* extend to :pep:`484`-compliant
      empty tuple type hints (i.e., ``typing.Tuple[()]``), which correctly
      preserve that subscripted empty tuple.

    **This getter lies less than the comparable**
    :func:`.get_hint_pep_typeargs_packed` **getter.** Whereas
    :func:`.get_hint_pep_typeargs_packed` synthetically propagates type variables from
    child to parent type hints (rather than preserving the literal type
    variables subscripting this type hint), this getter preserves the literal
    arguments subscripting this type hint if any. Notable cases where the two
    differ include:

    * Generic classes subclassing pseudo-superclasses subscripted by one or
      more type variables (e.g., ``class MuhGeneric(Generic[S, T])``).
    * Unions subscripted by one or more child type hints subscripted by one or
      more type variables (e.g., ``Union[str, Iterable[Tuple[S, T]]]``).

    Parameters
    ----------
    hint : object
        PEP-compliant type hint to be inspected.

    Returns
    -------
    tuple
        Either:

        * If this hint defines an ``__args__`` dunder attribute, the value of
          that attribute.
        * Else, the empty tuple.

    Raises
    ------
    BeartypeDecorHintPepException
        If this hint defines an ``__args__`` dunder attribute whose value is
        *not* a tuple.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_args)
       >>> get_hint_pep_args(typing.Any)
       ()
       >>> get_hint_pep_args(typing.List[int, str, typing.Dict[str, str]])
       (int, str, typing.Dict[str, str])
    '''

    # Tuple of the zero or more child type hints subscripting this hint if this
    # hint defines of the "__args__" dunder attribute *OR* "None" otherwise
    # (i.e., if this hint fails to define this attribute).
    hint_args = getattr(hint, '__args__', None)

    # If this hint does *NOT* define this attribute, return the empty tuple.
    if hint_args is None:
        return ()
    # Else, this hint defines this attribute.
    #
    # If this attribute is *NOT* a tuple...
    elif not isinstance(hint_args, tuple):
        # If this hint is the unsubscripted "typing.Union" hint semantically
        # equivalent to the subscripted "typing.Union[typing.Any]" hint, this
        # hint is a C-based type whose "__args__" dunder attribute is
        # implemented as a C-based slotted class attribute of some obscure type
        # under Python >= 3.14. Since unsubscripted "typing.Union" hints are
        # valid hints, this "__args__" implementation is *TECHNICALLY* also
        # valid albeit semantically meaningless. In this case, simply return the
        # empty tuple.
        if hint is Union:
            return ()
        # Else, this hint is *NOT* the unsubscripted "typing.Union" hint. In
        # this case, raise an exception.
        else:
            raise BeartypeDecorHintPepException(
                f'PEP-noncompliant hint {repr(hint)} '
                f'"__args__" dunder attribute {repr(hint_args)} '
                f'invalid (i.e., not tuple of child hints).'
            )
    # Else, this attribute is a tuple.
    # If this hint is subscripted by zero child type hints, this hint only
    # superficially appears to be unsubscripted but was actually subscripted by
    # the empty tuple (e.g., "tuple[()]", "typing.Tuple[()]"). Why? Because:
    # * Python 3.11 made the unfortunate decision of ambiguously conflating
    #   unsubscripted type hints (e.g., "tuple", "typing.Tuple") with type hints
    #   subscripted by the empty tuple, preventing downstream consumers from
    #   reliably distinguishing these two orthogonal cases.
    # * Python 3.9 made a similar decision but constrained to only PEP
    #   585-compliant empty tuple type hints (i.e., "tuple[()]"). PEP
    #   484-compliant empty tuple type hints (i.e., "typing.Tuple[()]")
    #   continued to correctly declare an "__args__" dunder attribute of "((),)"
    #   until Python 3.11.
    #
    # Disambiguate these two cases on behalf of callers by returning a 1-tuple
    # containing only the empty tuple (i.e., "((),)") rather than returning the
    # empty tuple (i.e., "()").
    elif not hint_args:
        return _HINT_ARGS_EMPTY_TUPLE
    # Else, this hint is either subscripted *OR* is unsubscripted but not PEP
    # 585-compliant.

    # In this case, return this tuple as is.
    return hint_args

# ....................{ GETTERS ~ typeargs                 }....................
#FIXME: Unit test us up, please. *sigh*
@callable_cached
def get_hint_pep_typeargs_unpacked(hint: Hint) -> (
    TuplePep484612646TypeArgsUnpacked):
    '''
    Tuple of all :pep:`484`-, :pep:`612`-, and :pep:`646`-compliant **unique
    unpacked type parameters** (i.e., :pep:`484`-compliant type variables,
    :pep:`612`-compliant unpacked parameter specifications, and
    :pep:`646`-compliant unpacked type variable tuples uniquely subscripting the
    passed PEP-compliant hint listed by the caller at hint declaration time,
    ignoring duplicates) if this hint is parametrized *or* the empty tuple
    otherwise (i.e., if this hint is unparametrized).

    This getter returns the tuple returned by the lower-level
    :func:`.get_hint_pep_typeargs_packed` getter, modified such that:

    * Each :pep:`646`-compliant type variable tuple (e.g., ``Ts`` such that
      ``Ts = typing.TypeVarTuple('Ts')``) is substituted with the equivalent
      unpacked type variable tuple (e.g., ``*Ts``).

    This getter is memoized for efficiency.

    Caveats
    -------
    **This getter should usually be called in lieu of calling the lower-level**
    :func:`.get_hint_pep_typeargs_packed` **getter.** Why? Because type
    parameters are *always* specified in unpacked rather than packed form.
    Packed type parameters are thus useless for most intents and purposes.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    TuplePep484612646TypeArgsUnpacked
        Either:

        * If this hint defines a ``__parameters__`` dunder attribute, the
          value of that attribute, modified such that:

          * Each :pep:`646`-compliant type variable tuple is substituted with
            the equivalent unpacked type variable tuple.

        * Else, the empty tuple.

    Raises
    ------
    BeartypeDecorHintPepException
        If this hint defines a ``__parameters__`` dunder attribute whose value
        is *not* a tuple.

    See Also
    --------
    :func:`.get_hint_pep_typeargs_packed`
        Further details.
    '''

    # Tuple of the zero or more type parameters parametrizing this hint if any
    # *OR* the empty tuple otherwise (i.e., if this hint is unparametrized).
    hint_typeargs = get_hint_pep_typeargs_packed(hint)

    # If this tuple is non-empty, this hint is parametrized. In this case...
    if hint_typeargs:
        # Avoid circular import dependencies.
        from beartype._util.hint.pep.proposal.pep646692 import (
            make_hint_pep646_typevartuple_unpacked_subbed)

        # List of all unpacked type parameters to be returned.
        hint_typeargs_unpacked = []

        # 0-based index of the current type parameter visited below.
        hint_typearg_index_curr = 0

        # 0-based index of the last type parameter to be visited below.
        hint_typearg_index_last = len(hint_typeargs) - 1

        # While the 0-based index of the currently visited type parameter does
        # *NOT* exceed that of the last such parameter...
        while hint_typearg_index_curr <= hint_typearg_index_last:
            # This possibly packed type parameter.
            hint_typearg = hint_typeargs[hint_typearg_index_curr]

            # If this is an undesirable PEP 646-compliant packed type variable
            # tuple, coerce this into a desirable PEP 646-compliant unpacked
            # type variable tuple.
            if isinstance(hint_typearg, HintPep646TypeVarTupleType):
                hint_typearg = make_hint_pep646_typevartuple_unpacked_subbed(
                    hint_typearg)
            # Else, this is *NOT* an undesirable PEP 646-compliant packed type
            # variable tuple. In this case, preserve this type parameter.
            #
            # In either case, this type parameter is now unpacked.

            # Append this now unpacked type parameter to this list.
            hint_typeargs_unpacked.append(hint_typearg)

            # Increment the index of the type parameter to be visited next.
            hint_typearg_index_curr += 1

        # Tuple of all unpacked type parameters to be returned, coerced from
        # this list.
        hint_typeargs = tuple(hint_typeargs_unpacked)

    # Return this tuple.
    return hint_typeargs  # type: ignore[return-value]


#FIXME: This getter *DEFINITELY* isn't behaving correctly at the moment -- but
#it's absolutely *NOT* our fault. It's CPython's and the "typing" module's
#fault. The issue is the standard "typing.Generic[...]" type hint factory, which
#is *NOT* correctly propagating type parameters onto subscripted generics: e.g.,
#    from beartype.typing import Generic, TypeVar
#    T = TypeVar('T')
#    class Ugh(Generic[T]): pass
#    print(get_hint_pep_typeargs_packed(Ugh))
#    print(get_hint_pep_typeargs_packed(Ugh[int]))
#
#The above incorrectly outputs:
#    (~T,)  # <-- this is good
#    ()  # <----- THIS IS BAD. wtf, "typing"?
#
#Oddly, PEP 695 resolved this subtle issue. When you use the more succinct
#syntax, get_hint_pep_typeargs_packed() behaves itself: e.g.,
#    from beartype.typing import Generic
#    class Ugh[T](): pass
#    print(get_hint_pep_typeargs_packed(Ugh))
#    print(get_hint_pep_typeargs_packed(Ugh[int]))
#
#The above incorrectly outputs:
#    (~T,)  # <-- this is good
#    (~T,)  # <-- this is good, too
#
#Thankfully, we're not sure anybody even cares about this. This has been an
#open issue for literally a decade until I finally discovered this madness in
#2025 Q2 while hacking on unrelated PEP 646 logic. Stupefying stuff. *sigh*
def get_hint_pep_typeargs_packed(hint: Hint) -> TuplePep484612646TypeArgsPacked:
    '''
    Tuple of all :pep:`484`-, :pep:`612`-, and :pep:`646`-compliant **unique
    packed type parameters** (i.e., :pep:`484`-compliant type variables,
    :pep:`612`-compliant parameter specifications, or :pep:`646`-compliant type
    variable tuples uniquely subscripting the passed PEP-compliant hint listed
    by the caller at hint declaration time, ignoring duplicates) if this hint is
    parametrized *or* the empty tuple otherwise (i.e., if this hint is
    unparametrized).

    This getter encapsulates the low-level ``__parameters__`` dunder attribute
    maintained by CPython and the :mod:`typing` module itself, intentionally
    returning similar results depending on whether the passed hint is a
    subscripted or unsubscripted generic. In any case, this getter returns a
    tuple compositing both:

    * **Direct parametrizations** (i.e., cases in which this hint itself is
      directly parametrized by type parameters).
    * **Superclass parametrizations** (i.e., cases in which this hint is
      indirectly parametrized by one or more superclasses of its class being
      directly parametrized by type parameters).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **The higher-level** :func:`.get_hint_pep_typeargs_unpacked` **getter should
    typically be called instead.** Why? Because type parameters are *always*
    specified in unpacked rather than packed form. Packed type parameters are
    thus useless for most intents and purposes. In fact, this getter only exists
    because the low-level ``__parameters__`` dunder attribute underlying this
    getter only lists type parameters in packed rather than unpacked form.

    **This getter should always be called in lieu of attempting to directly
    access the low-level** ``__parameters__`` **dunder attribute.** Various
    singleton objects defined by the :mod:`typing` module (e.g.,
    :attr:`typing.Any`, :attr:`typing.NoReturn`) fail to define this attribute,
    guaranteeing :class:`AttributeError` exceptions from all general-purpose
    logic attempting to directly access this attribute. Likewise, still other
    singleton objects defined by the :mod:`typing` module (e.g.,
    :attr:`typing.TypeAliasType`, :attr:`typing.Union`) define this attribute to
    be non-standard C-based class slots of some obscure type rather than
    standard tuples of type parameters. Thus this function, which "fills in the
    gaps" by circumventing these oversights.

    **Generics** (i.e., PEP-compliant type hints whose classes subclass one or
    more public :mod:`typing` pseudo-superclasses) are often but *not* always
    parametrized. For example, consider this unparametrized generic:

    .. code-block:: pycon

       >>> from typing import List
       >>> class UntypevaredGeneric(List[int]): pass
       >>> UntypevaredGeneric.__mro__
       (__main__.UntypevaredGeneric, list, typing.Generic, object)
       >>> UntypevaredGeneric.__parameters__
       ()

    Likewise, parametrized hints are often but *not* always generic. For example,
    consider this parametrized non-generic:

    .. code-block:: pycon

       >>> from typing import List, TypeVar
       >>> TypevaredNongeneric = List[TypeVar('T')]
       >>> type(TypevaredNongeneric).__mro__
       (typing._GenericAlias, typing._Final, object)
       >>> TypevaredNongeneric.__parameters__
       (~T,)

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    TuplePep484612646TypeArgsPacked
        Either:

        * If this hint defines a ``__parameters__`` dunder attribute, the
          value of that attribute.
        * Else, the empty tuple.

    Raises
    ------
    BeartypeDecorHintPepException
        If this hint defines a ``__parameters__`` dunder attribute whose value
        is *not* a tuple.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_typeargs_packed)

       >>> S = typing.TypeVar('S')
       >>> T = typing.TypeVar('T')
       >>> class UserList(typing.List[T]): pass

       >>> get_hint_pep_typeargs_packed(typing.Any)
       ()
       >>> get_hint_pep_typeargs_packed(typing.List[int])
       ()
       >>> get_hint_pep_typeargs_packed(typing.List[T])
       (T)
       >>> get_hint_pep_typeargs_packed(UserList)
       (T)
       >>> get_hint_pep_typeargs_packed(typing.List[T, int, S, str, T])
       (T, S)
    '''

    # Value of the "__parameters__" dunder attribute on this object if this
    # object defines this attribute (e.g., is *NOT* a PEP 585-compliant
    # unsubscripted generic) *OR* "None" otherwise (e.g., is such a generic).
    hint_typeargs = getattr(hint, '__parameters__', None)

    # If this object defines *NO* such attribute, synthetically reconstruct this
    # attribute for PEP 585-compliant unsubscripted generics. Notably...
    if hint_typeargs is None:
        # Reconstruct this attribute as either...
        hint_typeargs = (
            # If this hint is a PEP 585-compliant unsubscripted generic, the
            # tuple of all type variables parametrizing all pseudo-superclasses
            # of this generic;
            get_hint_pep585_generic_typeargs_packed(hint)
            if is_hint_pep585_generic_unsubbed(hint) else
            # Else, this hint is *NOT* a PEP 585-compliant unsubscripted
            # generic. In this case, the empty tuple.
            ()
        )
    # Else, this object defines this attribute.
    #
    # If this attribute is *NOT* a tuple...
    elif not isinstance(hint_typeargs, tuple):
        # If this hint is the unsubscripted "typing.Union" hint semantically
        # equivalent to the subscripted "typing.Union[typing.Any]" hint, this
        # hint is a C-based type whose "__parameters__" dunder attribute is
        # implemented as a C-based slotted class attribute of some obscure type
        # under Python >= 3.14. Since unsubscripted "typing.Union" hints are
        # valid hints, this "__parameters__" implementation is *TECHNICALLY*
        # also valid albeit semantically meaningless. In this case, simply
        # return the empty tuple.

        # If this hint is...
        if (
            # A type *AND*...
            isinstance(hint, type) and
            # This is a standard type well-known to violate PEP standards by
            # defining the "__parameters__" dunder attribute to *NOT* be a tuple
            # of PEP-compliant packed type parameters.
            hint in TYPES_NONPEP_TYPEARGS_PACKED
        ):
            # Then is a probably an unsubscripted standard type hint factory
            # (e.g., "typing.Union") semantically equivalent to a valid hint
            # (e.g., "typing.Union[typing.Any]"). This factory is probably
            # implemented as a C-based type whose "__parameters__" dunder
            # attribute is implemented as a C-based slotted class attribute of
            # some obscure type. Since this type is a valid hint, this otherwise
            # PEP-noncompliant "__parameters__" dunder attribute is technically
            # PEP-compliant (for certain broad definitions) despite being
            # semantically meaningless. Silently coerce this PEP-noncompliance
            # into PEP-compliance by reducing to the empty tuple.
            return ()
        # Else, this hint either is not a type *OR* is a user-defined type. In
        # either case, this hint defines an invalid "__parameters__" dunder
        # attribute and is thus PEP-noncompliant.

        # Raise an exception as a necessary fallback for safety.
        raise BeartypeDecorHintPepException(
            f'PEP-noncompliant hint {repr(hint)} '
            f'"__parameters__" dunder attribute value '
            f'{repr(hint_typeargs)} '
            f'invalid (i.e., not tuple of '
            f'PEP 484-, 612-, or PEP 646-compliant type parameters).'
        )

    # Return this attribute.
    return hint_typeargs

# ....................{ GETTERS ~ origin                   }....................
def get_hint_pep_origin(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    #FIXME: This should probably be a new "BeartypeDecorHintPepOriginException"
    #type, instead. But it's unclear whether users will even ever see this
    #exception. So, for now, laziness prevails. Huzzah! *sigh*
    exception_cls: TypeException = _BeartypeUtilTypeException,
    exception_prefix: str = '',
) -> Hint:
    '''
    **Unsafe origin object** (i.e., arbitrary object originating the passed
    PEP-compliant type hint but *not* necessarily an isinstanceable class such
    that all objects satisfying this hint are instances of this class)
    originating this hint if this hint originates from an origin *or*
    raise an exception otherwise (i.e., if this hint originates from *no*
    origin).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilTypeException`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    Hint
        Arbitrary object this hint originates from.

    Raises
    ------
    exception_cls
        If this hint originates from *no* object.
    '''

    # Origin originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_or_none(hint)

    # If this hint does *NOT* originate from anything, raise an exception.
    if hint_origin is None:
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'originates from no other object.'
        )
    # Else, this hint originates from something.

    # Return that something.
    return hint_origin


def get_hint_pep_origin_or_none(hint: Hint) -> HintOrNone:
    '''
    **Unsafe origin object** (i.e., arbitrary object originating the passed
    PEP-compliant type hint but *not* necessarily an isinstanceable class such
    that all objects satisfying this hint are instances of this class)
    originating this hint if this hint originates from an origin *or*
    :data:`None` otherwise (i.e., if this hint originates from *no* origin).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **The high-level** :func:`get_hint_pep_origin_type_isinstanceable` getter
    should always be called in lieu of this low-level function.** Whereas the
    former is guaranteed to return either an isinstanceable class or
    :data:`None`, this getter enjoys no such guarantees and instead returns an
    arbitrary object that may or may not actually be an instanceable class.

    If this getter *must* be called, **this getter should always be called in
    lieu of attempting to directly access the low-level** ``__origin__``
    **dunder attribute.** Various :mod:`typing` objects either fail to define
    this attribute or define this attribute non-orthogonally, including objects:

    * Failing to define this attribute altogether (e.g., :attr:`typing.Any`,
      :attr:`typing.NoReturn`).
    * Defining this attribute to be their unsubscripted :mod:`typing` type hint
      factories (e.g., :attr:`typing.Optional`, :attr:`typing.Union`).
    * Defining this attribute to be their actual origin types.

    Since the :mod:`typing` module neither guarantees the existence of this
    attribute nor imposes a uniform semantic on this attribute when defined,
    that attribute is *not* safely directly accessible. Thus this getter, which
    "fills in the gaps" by implementing this oversight.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    Optional[Hint]
        Either:

        * If this hint originates from an arbitrary object, that object.
        * Else, :data:`None`.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_origin_or_none)

       # This is sane.
       >>> get_hint_pep_origin_or_none(typing.List)
       list
       >>> get_hint_pep_origin_or_none(typing.List[int])
       list
       >>> get_hint_pep_origin_or_none(typing.Union)
       None
       >>> get_hint_pep_origin_or_none(typing.Union[int])
       None

       # This is insane.
       >>> get_hint_pep_origin_or_none(typing.Union[int, str])
       Union

       # This is crazy.
       >>> typing.Union.__origin__
       AttributeError: '_SpecialForm' object has no attribute '__origin__'

       # This is balls crazy.
       >>> typing.Union[int].__origin__
       AttributeError: type object 'int' has no attribute '__origin__'

       # This is balls cray-cray -- the ultimate evolution of crazy.
       >>> typing.Union[int, str].__origin__
       typing.Union
    '''

    # Return this hint's origin object if any *OR* "None" otherwise.
    return getattr(hint, '__origin__', None)

# ....................{ GETTERS ~ origin : type            }....................
#FIXME: Unit test us up, please.
def get_hint_pep_origin_type(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    #FIXME: This should probably be a new "BeartypeDecorHintPepOriginException"
    #type, instead. But it's unclear whether users will even ever see this
    #exception. So, for now, laziness prevails. Huzzah! *sigh*
    exception_cls: TypeException = _BeartypeUtilTypeException,
    exception_prefix: str = '',
) -> type:
    '''
    **Origin type** (i.e., class such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* raise an exception otherwise
    (i.e., if this hint originates from *no* such type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Type hint to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilTypeException`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    type
        Type originating this hint.

    Raises
    ------
    exception_cls
        If this hint either:

        * Does *not* originate from another object.
        * Originates from an object that is *not* a type.

    See Also
    --------
    :func:`.get_hint_pep_origin_or_none`
        Further details.
    '''

    # Origin type originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_type_or_none(hint)

    # If *NO* origin type originates this hint...
    if hint_origin is None:
        assert isinstance(exception_cls, type), (
            f'{exception_cls} not exception type.')
        assert isinstance(exception_prefix, str), (
            f'{exception_prefix} not string.')

        # Origin non-type originating this hint if any *OR* "None" otherwise.
        hint_origin_nontype = get_hint_pep_origin_or_none(hint)

        # If this hint does *NOT* originate from another object, raise an
        # appropriate exception.
        if hint_origin_nontype is None:
            raise exception_cls(
                f'{exception_prefix}type hint {repr(hint)} '
                f'originates from no other object.'
            )
        # Else, this hint originates from another object. By definition, this
        # object *CANNOT* be a type.

        # Raise an appropriate exception.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'originates from non-type {repr(hint_origin_nontype)}.'
        )
    # Else, this origin type originates this hint.

    # Return this origin type.
    return hint_origin


#FIXME: Unit test us up, please.
def get_hint_pep_origin_type_or_none(
    # Mandatory parameters.
    hint: Any,

    # Optional parameters.
    is_self_fallback: bool = False,
) -> Optional[type]:
    '''
    **Origin type** (i.e., class such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* :data:`None` otherwise (i.e.,
    if this hint originates from *no* such type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This high-level getter should always be called in lieu of either calling
    the low-level** :func:`.get_hint_pep_origin_or_none` **getter or attempting
    to directly access the low-level** ``__origin__`` **dunder attribute.**

    Parameters
    ----------
    hint : object
        Type hint to be inspected.
    is_self_fallback : bool = False
        :data:`True` only if returning the passed hint as a last-ditch fallback
        when this hint is a type defining the ``__origin__`` dunder attribute to
        be a non-type. Equivalently, if the passed hint is such a hint *and*
        this parameter is:

        * :data:`True`, this getter returns this hint as is.
        * :data:`False`, this getter returns :data:`None`.

        Defaults to :data:`False`. Explicit is better than implicit.

    Returns
    -------
    Optional[type]
        Either:

        * If this hint originates from a type, that type.
        * Else, :data:`None`.

    See Also
    --------
    :func:`.get_hint_pep_origin_or_none`
        Further details.
    '''
    assert isinstance(is_self_fallback, bool), (
        f'{repr(is_self_fallback)} not boolean.')

    # Unsafe origin type originating this hint if any *OR* "None" otherwise,
    # initialized to the arbitrary object set as the "hint.__origin__" dunder
    # attribute if this hint defines that attribute.
    hint_origin: Optional[type] = get_hint_pep_origin_or_none(hint)  # type: ignore[assignment]

    #FIXME: Unit test this up, please. *shrug*
    # If this origin is *NOT* a type...
    #
    # Ideally, this attribute would *ALWAYS* be a type for all possible
    # PEP-compliant type hints. For unknown reasons, type hint factories defined
    # by the standard "typing" module often set their origins to those same type
    # hint factories, despite those factories *NOT* being types. Why? Frankly,
    # we have no idea and neither does anyone else. Behold, true horror:
    #    >>> import typing
    #    >>> typing.Literal[1, 2].__origin__
    #    typing.Literal  # <-- do you even know what you are doing, python?
    #    >>> typing.Optional[int].__origin__
    #    typing.Union  # <-- wut? this is insane, python.
    if not isinstance(hint_origin, type):
        # Default this origin type to either...
        hint_origin = (
            # If...
            hint if (
                # The caller requests the "self" fallback logic *AND*...
                is_self_fallback and
                # This hint is itself a type, this hint could be euphemistically
                # said to originate from "itself." Fallback to this hint itself.
                # Look. Just go with it. We wave our hands in the air, fam.
                isinstance(hint, type)
            ) else
            # Else, either the caller did not request the "self" fallback logic
            # *OR* this hint is not a type. In either case, fallback to "None".
            None
        )
    # Else, this origin is a type.

    # Return this origin type.
    return hint_origin


def get_hint_pep_origin_type_isinstanceable(hint: Hint) -> type:
    '''
    **Isinstanceable origin type** (i.e., class passable as the second argument
    to the :func:`isinstance` builtin such that *all* objects satisfying the
    passed PEP-compliant type hint are instances of this class) originating
    this hint if this hint originates from such a type *or* raise an exception
    otherwise (i.e., if this hint does *not* originate from such a type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    type
        Standard origin type originating this hint.

    Raises
    ------
    BeartypeDecorHintPepException
        If this hint does *not* originate from a standard origin type.

    See Also
    --------
    :func:`get_hint_pep_origin_type_isinstanceable_or_none`
        Related getter.
    '''

    # Origin type originating this object if any *OR* "None" otherwise.
    hint_origin_type = get_hint_pep_origin_type_isinstanceable_or_none(hint)

    # If this type does *NOT* exist, raise an exception.
    if hint_origin_type is None:
        raise BeartypeDecorHintPepException(
            f'Type hint {repr(hint)} not isinstanceable (i.e., does not '
            f'originate from isinstanceable class).'
        )
    # Else, this type exists.

    # Return this type.
    return hint_origin_type


def get_hint_pep_origin_type_isinstanceable_or_none(
    hint: Hint) -> Optional[type]:
    '''
    **Standard origin type** (i.e., isinstanceable class declared by Python's
    standard library such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* :data:`None` otherwise (i.e.,
    if this hint does *not* originate from such a type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This high-level getter should always be called in lieu of the low-level**
    :func:`get_hint_pep_origin_or_none` **getter or attempting to
    directly access the low-level** ``__origin__`` **dunder attribute.**

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    Optional[type]
        Either:

        * If this hint originates from a standard origin type, that type.
        * Else, :data:`None`.

    See Also
    --------
    :func:`get_hint_pep_origin_type_isinstanceable`
        Related getter.
    :func:`get_hint_pep_origin_or_none`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Sign uniquely identifying this hint if any *OR* "None" otherwise.
    hint_sign = get_hint_pep_sign_or_none(hint)

    # Return either...
    return (
        # If this sign originates from an origin type, that type;
        get_hint_pep_origin_or_none(hint)  # type: ignore[return-value]
        if hint_sign in HINT_SIGNS_ORIGIN_ISINSTANCEABLE else
        # Else, "None".
        None
    )

# ....................{ PRIVATE ~ args                     }....................
#FIXME: Shift into the "beartype._data.hint" subpackage somewhere, please.
_HINT_ARGS_EMPTY_TUPLE = ((),)
'''
Tuple containing only the empty tuple, to be returned from the
:func:`.get_hint_pep_args` getter when passed either:

* A :pep:`585`-compliant type hint subscripted by the empty tuple (e.g.,
  ``tuple[()]``).
* A :pep:`484`-compliant type hint subscripted by the empty tuple (e.g.,
  ``typing.Tuple[()]``) under Python >= 3.11, which applied the :pep:`585`
  approach throughout the :mod:`typing` module.
'''



# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **ANSI utilities** (i.e., low-level callables handling ANSI escape
sequences colouring arbitrary strings).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import BoolTristate
from re import compile as re_compile

# ....................{ CONSTANTS                          }....................
ANSI_RESET = '\033[0m'
'''
ANSI escape sequence resetting the effect of all prior ANSI sequence sequences,
effectively "undoing" all colors and styles applied by those sequences.
'''

# ....................{ CONSTANTS ~ color                  }....................
COLOR_BLUE = '\033[34m'
'''
ANSI escape sequence colouring all subsequent characters as blue.
'''


COLOR_CYAN = '\033[36m'
'''
ANSI escape sequence colouring all subsequent characters as **cyan** (i.e.,
light blue).
'''


COLOR_GREEN = '\033[32m'
'''
ANSI escape sequence colouring all subsequent characters as **green**.
'''


COLOR_MAGENTA = '\033[35m'
'''
ANSI escape sequence colouring all subsequent characters as **magenta** (i.e.,
purple, dark blue).
'''


COLOR_RED = '\033[31m'
'''
ANSI escape sequence colouring all subsequent characters as **red**.
'''


COLOR_YELLOW = '\033[33m'
'''
ANSI escape sequence colouring all subsequent characters as **yellow**.
'''

# ....................{ CONSTANTS ~ style                  }....................
STYLE_BOLD = '\033[1m'
'''
ANSI escape sequence stylizing all subsequent characters as bold.
'''

# ....................{ TESTERS                            }....................
def is_str_ansi(text: str) -> bool:
    '''
    :data:`True` only if the passed text contains one or more ANSI escape
    sequences.

    Parameters
    ----------
    text : str
        Text to be tested.

    Returns
    -------
    bool
        :data:`True` only if this text contains one or more ANSI escape
        sequences.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return true only this compiled regex matching ANSI escape sequences
    # returns a non-"None" match object when passed this text.
    return _ANSI_REGEX.search(text) is not None

# ....................{ COLOURIZERS                        }....................
def color_hint(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = True,
) -> str:
    '''
    Colour the passed substring as a type hint if the passed tri-state colouring
    boolean instructs this function to do so.

    Parameters
    ----------
    text : str
        Text to be coloured as a type hint.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`True`.

    Returns
    -------
    str
        This text conditionally coloured as a type hint.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to colour this
        # string, this string coloured with ANSI;
        f'{STYLE_BOLD}{COLOR_GREEN}{text}{ANSI_RESET}'
        if _is_color(is_color) else
        # Else, this string uncoloured.
        text
    )


def color_pith(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = True,
) -> str:
    '''
    Colour the passed substring as a **pith representation** (i.e.,
    machine-readable string describing the value of the object currently being
    type-checked, typically created by the
    :func:`beartype._util.text.utiltextrepr.represent_object` function) if the
    passed tri-state colouring boolean instructs this function to do so.

    Parameters
    ----------
    text : str
        Text to be coloured as a representation.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`True`.

    Returns
    -------
    str
        This text conditionally coloured as a representation.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to colour this
        # string, this string coloured with ANSI;
        f'{STYLE_BOLD}{COLOR_RED}{text}{ANSI_RESET}'
        if _is_color(is_color) else
        # Else, this string uncoloured.
        text
    )


def color_type(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = True,
) -> str:
    '''
    Colour the passed substring as a simple class if the passed tri-state
    colouring boolean instructs this function to do so.

    Parameters
    ----------
    text : str
        Text to be coloured as a simple class.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`True`.

    Returns
    -------
    str
        This text conditionally coloured as a simple class.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to colour this
        # string, this string coloured with ANSI;
        f'{STYLE_BOLD}{COLOR_YELLOW}{text}{ANSI_RESET}'
        if _is_color(is_color) else
        # Else, this string uncoloured.
        text
    )

# ....................{ COLOURIZERS ~ name                 }....................
def color_attr_name(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = True,
) -> str:
    '''
    Colour the passed substring as a **Python identifier** (e.g., possibly
    fully-qualified name of a module, class, callable, or variable name) if the
    passed tri-state colouring boolean instructs this function to do so.

    Parameters
    ----------
    text : str
        Text to be coloured as a Python identifier.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`True`.

    Returns
    -------
    str
        This text conditionally coloured as a Python identifier.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to colour this
        # string, this string coloured with ANSI;
        f'{STYLE_BOLD}{COLOR_MAGENTA}{text}{ANSI_RESET}'
        if _is_color(is_color) else
        # Else, this string uncoloured.
        text
    )


def color_arg_name(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = True,
) -> str:
    '''
    Colour the passed substring as an **argument name** (i.e., of the parameter
    of a :func:`beartype.beartype`-decorated callable currently being
    type-checked) if the passed tri-state colouring boolean instructs this
    function to do so.

    Parameters
    ----------
    text : str
        Text to be coloured as an argument name.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`True`.

    Returns
    -------
    str
        This text coloured as an argument name.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to colour this
        # string, this string coloured with ANSI;
        f'{STYLE_BOLD}{COLOR_BLUE}{text}{ANSI_RESET}'
        if _is_color(is_color) else
        # Else, this string uncoloured.
        text
    )

# ....................{ STRIPPERS                          }....................
#FIXME: Unit test up the "is_color" parameter.
def strip_str_ansi(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Strip *all* ANSI escape sequences from the passed string if the passed
    tri-state colouring boolean instructs this function to do so.

    Specifically:

    * If ``is_color is True``, this function silently reduces to a noop.
    * If ``is_color is False``, this function unconditionally strips all ANSI
      escape sequences from this string.
    * If ``is_color is None``, this function conditionally strips all ANSI
      escape sequences from this string only if standard output is currently
      attached to an interactive terminal.

    Parameters
    ----------
    text : str
        Text to be stripped.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        This text conditionally stripped of ANSI.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Return either...
    return (
        # If this tri-state boolean instructs this function to preserve all ANSI
        # in this string, this string unmodified;
        text
        if _is_color(is_color) else
        # Else, this string stripped of all ANSI.
        _ANSI_REGEX.sub('', text)
    )

# ....................{ PRIVATE ~ constants                }....................
_ANSI_REGEX = re_compile(r'\033\[[0-9;?]*[A-Za-z]')
'''
Compiled regular expression matching a single ANSI escape sequence.
'''

# ....................{ PRIVATE ~ testers                  }....................
def _is_color(is_color: BoolTristate) -> bool:
    '''
    Reduce the passed tri-state colouring boolean governing ANSI usage to a
    simple boolean.

    Specifically, this tester returns either:

    * :data:`True` only if the passed ``is_color`` parameter is either:

      * :data:`True`.
      * :data:`None` and standard output is currently attached to an interactive
        POSIX-compliant terminal. Note that this is the common case, as the
        :attr:`beartype.BeartypeConf.is_color` attribute underlying this
        parameter typically defaults to :data:`None`.

    * :data:`False` otherwise.

    Parameters
    ----------
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
    '''
    assert isinstance(is_color, bool) or is_color is None, (  # <-- "NoneTypeOr" is unavailable here
        f'{repr(is_color)} not tri-state boolean.')

    # Avoid circular import dependencies.
    from beartype._util.os.utilostty import is_stdout_terminal

    # Return true only if the passed tri-state boolean is either...
    return (
        # True *OR*...
        is_color is True or
        # "None" and standard output is currently attached to an interactive
        # POSIX-compliant terminal.
        (is_color is None and is_stdout_terminal())
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Poetry integration tests.**

This submodule functionally tests that the third-party Poetry packaging devtool
successfully installs this project.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('poetry')
def test_poetry(
    monkeypatch: 'pytest.MonkeyPatch', tmp_path: 'pathlib.Path') -> None:
    '''
    Integration test testing that the third-party Poetry packaging devtool
    successfully installs this package into a placeholder package defined by a
    minimal-length ``pyproject.toml`` file requiring this package as a mandatory
    dependency.

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        :mod:`pytest` fixture allowing various state associated with the active
        Python process to be temporarily changed for the duration of this test.
    tmp_path : pathlib.Path
        Abstract path encapsulating a temporary directory unique to this test,
        created in the base temporary directory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.path.utilpathcopy import copy_dir
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import (
        run_command_forward_output)
    from beartype_test._util.path.pytpathmain import get_main_dir
    from beartype_test._util.path.pytpathtest import (
        get_test_func_data_lib_poetry_dir)

    # ....................{ LOCALS                         }....................
    # Path object encapsulating the absolute dirname of the top-level directory
    # providing the entirety of beartype.
    beartype_dir_src = get_main_dir()

    # Path object encapsulating the absolute dirname of this same directory
    # symbolically linked into this temporary directory under the dirname
    # referenced by the "pyproject.toml" file for this Poetry project:
    #     # In that "pyproject.toml" file:
    #     beartype = {path = "../beartype", develop = true}
    beartype_dir_trg = tmp_path / 'beartype'

    # Path object encapsulating the absolute dirname of the test-specific data
    # directory containing an empty Poetry-managed project whose
    # "pyproject.toml" file requires "beartype" as a Poetry-specific mandatory
    # runtime dependency.
    project_dir_src = get_test_func_data_lib_poetry_dir()

    # Path object encapsulating the absolute dirname of this same Poetry project
    # copied into this temporary directory. Although a symbolic link *MIGHT*
    # suffice as well, we lack in Poetry. In all likelihood, Poetry expects to
    # be able to safely modify the contents of this directory.
    project_dir_trg = tmp_path / 'poetry'

    # ....................{ PATHS                          }....................
    # Symbolically link the beartype directory into this temporary directory.
    beartype_dir_trg.symlink_to(beartype_dir_src, target_is_directory=True)

    # Recursively copy this Poetry project into this temporary directory.
    copy_dir(src_dirname=project_dir_src, trg_dirname=project_dir_trg)

    # Change the current working directory (CWD) to that of this Poetry project.
    monkeypatch.chdir(project_dir_trg)

    # ....................{ COMMANDS                       }....................
    # # Tuple of all shell words with which to run the external "poetry" command.
    POETRY_ARGS = get_interpreter_command_words() + (
        # Fully-qualified name of the "poetry" package to be run.
        '-m', 'poetry',

        # Instruct Poetry to run the "install" subcommand.
        'install',

        # Force Poetry to implicitly answer "Yes" to all interactive queries.
        '--no-interaction',

        # *ONLY* install beartype as a mandatory runtime dependency of this
        # Poetry project. Do *NOT* bother installing this Poetry project itself.
        '--no-root',

        #FIXME: Let's see if this suffices. If so, wonderful! If not, we sigh.
        '--dry-run',
    )

    # Run the "poetry" command in the current ${PATH} with these options and
    # arguments, raising an exception on subprocess failure while forwarding
    # all standard output and error output by this subprocess to the
    # standard output and error file handles of the active Python process.
    run_command_forward_output(command_words=POETRY_ARGS)
